/*
 *
 *  (c) COPYRIGHT INRIA and W3C, 1996-2003
 *  Please first read the full copyright statement in file COPYRIGHT.
 *
 */

/*
 *
 * Authors: V. Quint
 *          I. Vatton - new version of table management
 */

/* Amaya includes  */
#define THOT_EXPORT extern
#include "amaya.h"
#include "fetchXMLname_f.h"
#include "undo.h"
#include "MathML.h"
#include "HTMLedit_f.h"
#include "HTMLtable_f.h"
#include "MathMLbuilder_f.h"
#include "Mathedit_f.h"

static Element      CurrentRow = NULL;
static Element      CurrentColumn = NULL;
static Element      CurrentCell = NULL;
static int          PreviousColspan;
static int          PreviousRowspan;
static ThotBool     NewTable = FALSE;

/*----------------------------------------------------------------------
  GetSiblingRow
  returns the sibling row before or after the cuttent row.
  ----------------------------------------------------------------------*/
static Element GetSiblingRow (Element row, ThotBool before, ThotBool inMath)
{
  ElementType         elType;
  SSchema             rowSS;

  if (row)
    {
      rowSS = TtaGetElementType(row).ElSSchema;
      do
	{
	  if (before)
	    TtaPreviousSibling (&row);
	  else
	    TtaNextSibling (&row);
	  elType = TtaGetElementType (row);
	  if (elType.ElSSchema == rowSS &&
	      (inMath && (elType.ElTypeNum == MathML_EL_MTR ||
			  elType.ElTypeNum == MathML_EL_MLABELEDTR)) ||
	      (!inMath && elType.ElTypeNum == HTML_EL_Table_row))
	    return row;
	} 
      while (row);
    }
  return row;
}

/*----------------------------------------------------------------------
  GetSiblingCell
  returns the sibling cell before or after the cuttent cell.
  ----------------------------------------------------------------------*/
static Element GetSiblingCell (Element cell, ThotBool before, ThotBool inMath)
{
  ElementType         elType;
  SSchema             cellSS;

  if (cell)
    {
      cellSS = TtaGetElementType(cell).ElSSchema;
      do
	{
	  if (before)
	    TtaPreviousSibling (&cell);
	  else
	    TtaNextSibling (&cell);
	  elType = TtaGetElementType (cell);
	  if (elType.ElSSchema == cellSS &&
	      ((inMath && elType.ElTypeNum == MathML_EL_MTD) ||
	       (!inMath && (elType.ElTypeNum == HTML_EL_Data_cell ||
			    elType.ElTypeNum == HTML_EL_Heading_cell))))
	    return cell;
	} 
      while (cell);
    }
  return cell;
}

/*----------------------------------------------------------------------
   SetRowExt
   Set the attribute RowExt of cell "cell" in row "row" according to span.
  ----------------------------------------------------------------------*/
int SetRowExt (Element cell, int span, Document doc, ThotBool inMath)
{
   Element             row, spannedrow, nextspannedrow;
   ElementType         elType;
   AttributeType       attrType;
   Attribute           attr;
   int                 overflow;

   overflow = 0;
   row = TtaGetParent (cell);
   elType = TtaGetElementType (row);
   attrType.AttrSSchema = elType.ElSSchema;
   if (inMath)
     attrType.AttrTypeNum = MathML_ATTR_MRowExt;
   else
     attrType.AttrTypeNum = HTML_ATTR_RowExt;

   if (span == 1 || span < 0)
     /* remove attribute RowExt if it is attached to the cell */
     {
       attr = TtaGetAttribute (cell, attrType);
       if (attr)
	 TtaRemoveAttribute (cell, attr, doc);
     }
   else
     {
       if (span == 0)
	 span = THOT_MAXINT;
       spannedrow = row;
       while (span > 0 && spannedrow != NULL)
	 {
	   span--;
	   nextspannedrow = GetSiblingRow (spannedrow, FALSE, inMath);
	   if ((span == 0 || nextspannedrow == NULL) && spannedrow != row)
	     {
	       attr = TtaGetAttribute (cell, attrType);
	       if (attr == NULL)
		 {
		   attr = TtaNewAttribute (attrType);
		   if (attr != NULL)
		     TtaAttachAttribute (cell, attr, doc);
		 }
	       if (attr)
		 TtaSetAttributeReference (attr, cell, doc, spannedrow, doc);
	       overflow = span;
	     }
	   spannedrow = nextspannedrow;
	 }
     }
   return overflow;
}

/*----------------------------------------------------------------------
   SetColExt
   Sets the attribute ColExt of cell "cell" according to span.
   If the value of span is greater than the number of available cells,
   return this difference.
  ----------------------------------------------------------------------*/
int SetColExt (Element cell, int span, Document doc, ThotBool inMath,
		ThotBool deletingLastCol)
{
  ElementType         elType;
  Element             colHead, nextColHead, prevColHead;
  AttributeType       attrType, refColType;
  Attribute           attr, refColAttr;
  int                 i, overflow;
  char                name[50];
  Document            refdoc;

  overflow = 0;
  elType = TtaGetElementType (cell);
  attrType.AttrSSchema = elType.ElSSchema;
  refColType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    {
      attrType.AttrTypeNum = MathML_ATTR_MColExt;
      refColType.AttrTypeNum = MathML_ATTR_MRef_column;
    }
  else
    {
      attrType.AttrTypeNum = HTML_ATTR_ColExt;
      refColType.AttrTypeNum = HTML_ATTR_Ref_column;
    }
  if (span == 1 || span < 0)
    /* remove attribute ColExt */
    {
      attr = TtaGetAttribute (cell, attrType);
      if (attr)
	TtaRemoveAttribute (cell, attr, doc);
    }
  else
    {
      /* set attribute ColExt */
      attr = TtaGetAttribute (cell, attrType);
      if (!attr)
	{
	  attr = TtaNewAttribute (attrType);
	  if (attr)
	    TtaAttachAttribute (cell, attr, doc);
	}
      if (attr)
	{
	  refColAttr = TtaGetAttribute (cell, refColType);
	  if (refColAttr)
	    {
	      TtaGiveReferenceAttributeValue (refColAttr, &colHead, name,
					      &refdoc);
	      if (colHead)
		{
		  nextColHead = colHead;
		  prevColHead = NULL;
		  for (i = 0; (i < span || span == 0) && nextColHead; i++)
		    {
		      colHead = nextColHead;
		      TtaNextSibling (&nextColHead);
		      if (nextColHead)
			prevColHead = colHead;
		    }
		  if (i < span && span != 0)
		    /* the value of span is wrong. It's greater than the
		       number of columns */
		    overflow = span - i;
		  if (deletingLastCol && span == 0)
		    colHead = prevColHead;
		  TtaSetAttributeReference (attr, cell, doc, colHead, doc);
		}
	    }
	}
    }
  return overflow;
}

/*----------------------------------------------------------------------
   GetCellFromColumnHead
   returns the cell that corresponds to the Column_head element colhead
   in a given row.
  ----------------------------------------------------------------------*/
static Element GetCellFromColumnHead (Element row, Element colhead,
				      ThotBool inMath)
{
   Element             cell, currentcolhead;
   ThotBool            found;
   ElementType         elType, cellType;
   AttributeType       attrType;
   Attribute           attr;
   char                name[50];
   Document            refdoc;

   cell = NULL;
   /* if it's not a row, return immediately with NULL */
   elType = TtaGetElementType (row);
   if ((inMath && (elType.ElTypeNum == MathML_EL_MTR ||
		   elType.ElTypeNum == MathML_EL_MLABELEDTR)) ||
       (!inMath && elType.ElTypeNum == HTML_EL_Table_row))
     {
       attrType.AttrSSchema = elType.ElSSchema;
       if (inMath)
	 attrType.AttrTypeNum = MathML_ATTR_MRef_column;
       else
	 attrType.AttrTypeNum = HTML_ATTR_Ref_column;

       cell = TtaGetFirstChild (row);
       found = FALSE;
       while (cell != NULL && !found)
	 {
	   cellType = TtaGetElementType (cell);
	   if (cellType.ElSSchema == elType.ElSSchema &&
	       ((inMath && cellType.ElTypeNum == MathML_EL_MTD) ||
		(!inMath && (cellType.ElTypeNum == HTML_EL_Data_cell ||
			     cellType.ElTypeNum == HTML_EL_Heading_cell))))
	     {
	       attr = TtaGetAttribute (cell, attrType);
	       if (attr != NULL)
		 {
		   TtaGiveReferenceAttributeValue (attr, &currentcolhead, name,
						   &refdoc);
		   if (currentcolhead == colhead)
		     found = TRUE;
		 }
	     }
	   if (!found)
	     TtaNextSibling (&cell);
	 }
     }
   return cell;
}

/*----------------------------------------------------------------------
   GetCloseCell
   returns the cell that corresponds to the Column_head element colhead
   or a previous or next colhead in a given row.
   - When the parameter before is TRUE, a previous cell is returned.
     The parameter spanupdate returns TRUE if the previous cell is extended
     (add = TRUE) or reduced (add = FALSE) and the parameter rowspan
     returns its rowspan value (1 by default).
   - When the parameter before is FALSE, the next cell is returned.
  ----------------------------------------------------------------------*/
static Element GetCloseCell (Element row, Element colhead,
			     Document doc, ThotBool before, ThotBool add,
			     ThotBool inMath, ThotBool *spanupdate,
			     int *rowspan, ThotBool deletingLastCol)
{
  Element             col, cell;
  ElementType         elType;
  Attribute           attr;
  AttributeType       attrType, attrTypeRef;
  int                 colspan, pos;

  *spanupdate = FALSE;
  *rowspan = 1;
  col = colhead;
  elType = TtaGetElementType (col);
  cell = GetCellFromColumnHead (row, colhead, inMath);
  pos = 0;
  while (!cell && col)
    {
      /* no cell related to this column in this row */
      if (before)
	TtaPreviousSibling (&col);
      else
	TtaNextSibling (&col);
      cell = GetCellFromColumnHead (row, col, inMath);
      pos++;
    }

  if (cell && before)
    {
      /* get the colspan value of the element */
      attrType.AttrSSchema = elType.ElSSchema;
      attrTypeRef.AttrSSchema = elType.ElSSchema;
      if (inMath)
	{
	  attrTypeRef.AttrTypeNum = MathML_ATTR_MColExt;
	  attrType.AttrTypeNum = MathML_ATTR_columnspan;
	}
      else
	{
	  attrTypeRef.AttrTypeNum = HTML_ATTR_ColExt;
	  attrType.AttrTypeNum = HTML_ATTR_colspan_;
	}
      attr = TtaGetAttribute (cell, attrType);
      if (attr)
	{
	  colspan = TtaGetAttributeValue (attr);
	  if (colspan < 0)
	    colspan = 1;
	  if (colspan == 0 ||  /* 0 means infinite span */
	      (add && colspan - pos > 1) ||
	      (!add && colspan - pos >= 1))
	    {
	      if (colspan > 0)
		if (!add && colspan == 2)
		  {
		    colspan--;
		    TtaRegisterAttributeDelete (attr, cell, doc);
		    TtaRemoveAttribute (cell, attr, doc);
		    attr = TtaGetAttribute (cell, attrTypeRef);
		    if (attr)
		      TtaRemoveAttribute (cell, attr, doc);
		  }
		else
		  {
		    if (add)
		      colspan++;
		    else
		      colspan--;
		    if (!add)
		      TtaRegisterAttributeReplace (attr, cell, doc);
		    TtaSetAttributeValue (attr, colspan, cell, doc);
		  }
	      if ((!add && colspan > 0 && colspan - pos == 0) ||
		  colspan == 0)
		SetColExt (cell, colspan, doc, inMath, deletingLastCol);
	      *spanupdate = TRUE;
	      /* check its rowspan attribute */
	      if (inMath)
		attrType.AttrTypeNum = MathML_ATTR_rowspan_;
	      else
		attrType.AttrTypeNum = HTML_ATTR_rowspan_;
	      attr = TtaGetAttribute (cell, attrType);
	      if (attr)
		/* this cell has an attribute rowspan */
		*rowspan = TtaGetAttributeValue (attr);
	      if (*rowspan < 0)
		*rowspan = 1;
	    }
	}
    }
  return (cell);
}

/*----------------------------------------------------------------------
   LinkCellToColumnHead link a cell to the Column_head representing the
   column it belongs to.
  ----------------------------------------------------------------------*/
static void LinkCellToColumnHead (Element cell, Element colhead,
				      Document doc, ThotBool inMath)
{
  ElementType         elType;
  AttributeType       attrType;
  Attribute           attr;

  if (cell == NULL || colhead == NULL)
    return;
  elType = TtaGetElementType (cell);
  attrType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    attrType.AttrTypeNum = MathML_ATTR_MRef_column;
  else
    attrType.AttrTypeNum = HTML_ATTR_Ref_column;
  attr = TtaGetAttribute (cell, attrType);
  if (attr == NULL)
    {
      attr = TtaNewAttribute (attrType);
      if (attr)
	TtaAttachAttribute (cell, attr, doc);
    }
  if (attr)
    TtaSetAttributeReference (attr, cell, doc, colhead, doc);
}

/*----------------------------------------------------------------------
   AddEmptyCellInRow add an empty cell element in the given row and relates
   it to the given Column_head colhead.
   The parameter sibling gives the cell which will precede or follow the new
   cell according to the before value.
   withUndo indicate whether the new empty cell must be registered in the
   undo queue or not.
   Return the created empty cell.
  ----------------------------------------------------------------------*/
static Element AddEmptyCellInRow (Element row, Element colhead,
				  Element sibling, ThotBool before,
				  Document doc, ThotBool inMath,
				  ThotBool placeholder, ThotBool withUndo)
{
  Element             lastcell, constr;
  ElementType         elType;
  Attribute           attr;
  AttributeType       attrType;

  lastcell = NULL;
  if (row == NULL)
    return (NULL);

  elType = TtaGetElementType (row);
  if (inMath)
    elType.ElTypeNum = MathML_EL_MTD;
  else
    elType.ElTypeNum = HTML_EL_Data_cell;
  lastcell = TtaNewTree (doc, elType, "");
  if (lastcell)
    {
      if (sibling)
	TtaInsertSibling (lastcell, sibling, before, doc);
      else
	TtaInsertFirstChild (&lastcell, row, doc);
      LinkCellToColumnHead (lastcell, colhead, doc, inMath);
      if (inMath && placeholder)
	/* put attribute IntPlaceholder on the empty Construct that has
	   been created in the new cell */
	{
	  elType.ElTypeNum = MathML_EL_Construct;
	  constr = TtaSearchTypedElement (elType, SearchInTree, lastcell);
	  if (constr)
	    {
	      attrType.AttrSSchema = elType.ElSSchema;
	      attrType.AttrTypeNum = MathML_ATTR_IntPlaceholder;
	      attr = TtaNewAttribute (attrType);
	      if (attr)
		{
		  TtaAttachAttribute (constr, attr, doc);
		  TtaSetAttributeValue (attr,
			   MathML_ATTR_IntPlaceholder_VAL_yes_, constr, doc);
		}
	    }
	}
      if (withUndo)
	TtaRegisterElementCreate (lastcell, doc);
    }
  return (lastcell);
}

/*----------------------------------------------------------------------
  CreateCellsInRowGroup
  ----------------------------------------------------------------------*/
static void CreateCellsInRowGroup (Element group, Element prevCol,
				   Element nextCol, Element colhead,
				   Document doc, ThotBool inMath,
				   ThotBool last)
{
  Element          row, child;
  ElementType      elType;
  int              rowspan;
  ThotBool         span;

  row = TtaGetFirstChild (group);
  /* skip the first elements that are now rows */
  if (row)
    elType = TtaGetElementType (row);
  while (row && elType.ElTypeNum != HTML_EL_Table_row)
    {
      TtaNextSibling (&row);
      if (row)
	elType = TtaGetElementType (row);
    }
  /* process all rows in this group */
  while (row)
    {
      span = FALSE;
      rowspan = 1;
      /* get the sibling cell */
      if (last)
	child = TtaGetLastChild (row);
      else
	/* look for the previous cell */
	child = GetCloseCell (row, prevCol, doc, TRUE, TRUE, inMath,
			      &span, &rowspan, FALSE);
      if (child)
	{
	  if (!span)
	    /* add a new cell after */
	    AddEmptyCellInRow (row, colhead, child, FALSE, doc,
			       inMath, FALSE, TRUE);
	}
      else
	{
	  /* look for the next cell */
	  child = GetCloseCell (row, nextCol, doc, FALSE, TRUE, inMath,
				&span, &rowspan, FALSE);
	  /* add before */
	  AddEmptyCellInRow (row, colhead, child, TRUE, doc, inMath,
			     FALSE, TRUE);
	}
      if (rowspan == 0)
	rowspan = THOT_MAXINT;
      while (rowspan >= 1 && row)
	{
	  row = GetSiblingRow (row, FALSE, inMath);
	  rowspan--;
	}
    }
}

/*----------------------------------------------------------------------
  NewColumnHead creates a new Column_head and returns it.   
  If generateEmptyCells == TRUE, create an additional empty cell in all
  rows, except the row indicated.
  If last == TRUE when lastcolhead is the last current column.
  The parameter before indicates if the lastcolhead precedes or follows
  the new created Column_head. It should be FALSE when last is TRUE.
  ----------------------------------------------------------------------*/
Element NewColumnHead (Element lastcolhead, ThotBool before,
		       ThotBool last, Element row, Document doc,
		       ThotBool inMath, ThotBool generateEmptyCells)
{
  Element             colhead, currentrow;
  Element             group, groupdone, block;
  Element             prevCol, nextCol;
  Element             table, child;
  ElementType         elType;
  int                 rowspan;
  ThotBool            select, backward, span;

  if (lastcolhead == NULL)
    return NULL;
  select = (row == NULL);
  span = FALSE;
  elType = TtaGetElementType (lastcolhead);
  colhead = TtaNewTree (doc, elType, "");
  if (colhead != NULL)
    {
      elType.ElTypeNum = HTML_EL_Table;
      table = TtaGetTypedAncestor (lastcolhead, elType);
      prevCol = nextCol = lastcolhead;
      if (before)
	TtaPreviousSibling (&prevCol);
      else
	TtaNextSibling (&nextCol);	

      TtaInsertSibling (colhead, lastcolhead, before, doc);
      TtaRegisterElementCreate (colhead, doc);
      if (generateEmptyCells)
	/* add empty cells to all other rows */
	{
	  if (row)
	    {
	      /* process the row group except the given row */
	      currentrow = GetSiblingRow (row, TRUE, inMath);
	      if (!last && currentrow == NULL)
		{
		  /* when last is TRUE, only cells of previous rows
		     should be created  */
		  currentrow = GetSiblingRow (row, FALSE, inMath);
		  backward = FALSE;
		}
	      else
		backward = TRUE;
	    }
	  else
	    {
	      elType = TtaGetElementType (lastcolhead);
	      /* get the first row */
	      if (inMath)
		elType.ElTypeNum = MathML_EL_TableRow;
	      else
		elType.ElTypeNum = HTML_EL_Table_row;
	      currentrow = TtaSearchTypedElement (elType, SearchInTree,
						  table);
	      backward = FALSE;
	    }

	  while (currentrow)
	    {
	      rowspan = 1;
	      /* get the sibling cell */
	      if (last)
		child = TtaGetLastChild (currentrow);
	      else
		/* look for the previous cell */
		child = GetCloseCell (currentrow, prevCol, doc, TRUE,
				      TRUE, inMath, &span, &rowspan, FALSE);
	      if (child)
		{
		  if (!span)
		    {
		      if (select && row == NULL)
			/* first row where a cell is created */
			row = currentrow;
		      /* add a new cell after */
		      AddEmptyCellInRow (currentrow, colhead, child, FALSE,
					 doc, inMath, FALSE, TRUE);
		      TtaChangeInfoLastRegisteredElem (doc, 3);
		    }
		}
	      else
		{
		  /* look for the next cell */
		  child = GetCloseCell (currentrow, nextCol, doc, FALSE,
					TRUE, inMath, &span, &rowspan, FALSE);
		  if (select && row == NULL)
		    /* first row where a cell is created */
		    row = currentrow;
		  /* add a cell before */
		  AddEmptyCellInRow (currentrow, colhead, child, TRUE, doc,
				     inMath, FALSE, TRUE);
		  TtaChangeInfoLastRegisteredElem (doc, 3);
		}
	      if (rowspan == 0)
		rowspan = THOT_MAXINT;
	      while (rowspan >= 1 && currentrow)
		{
		  if (backward)
		    {
		      TtaPreviousSibling (&currentrow);
		      if (currentrow == NULL && !last)
			{
			  /* we have to manage following rows too */
			  currentrow = GetSiblingRow (row, FALSE, inMath);
			  backward = FALSE;
			  rowspan = 1;
			}
		    }
		  else
		    currentrow = GetSiblingRow (currentrow, FALSE, inMath);
		  rowspan--;
		}
	    }

	  if (!inMath)
	    {
	      groupdone = TtaGetParent (row);	/* done with this group */
	      /* process the other row groups */
	      /* visit all children of the Table element */
	      block = TtaGetFirstChild (table);
	      while (block)
		{
		  elType = TtaGetElementType (block);
		  if (elType.ElTypeNum == HTML_EL_thead ||
		      elType.ElTypeNum == HTML_EL_tfoot)
		    {
		      /* this block is a thead or tfoot element */
		      group = block;
		      if (group != groupdone)
			CreateCellsInRowGroup (group, prevCol, nextCol,
					       colhead, doc, inMath, last);
		      else if (last)
			block = NULL;
		    }
		  else if (elType.ElTypeNum == HTML_EL_Table_body)
		    {
		      /* this child is the Table_body element */
		      /* get the first tbody element */
		      group = TtaGetFirstChild (block);
		      /* process all tbody elements */
		      while (group != NULL)
			{
			  if (group != groupdone)
			    CreateCellsInRowGroup (group, prevCol, nextCol,
						   colhead, doc, inMath, last);
			  if (last)
			    {
			      group = NULL;
			      block = NULL;
			    }
			  else
			    TtaNextSibling (&group);
			}
		    }
		  if (block)
		    TtaNextSibling (&block);
		}
	    }
	}
    }
  if (select)
    {
      child = GetCellFromColumnHead (row, colhead, inMath);
      child = TtaGetFirstChild (child);
      if (child)
	TtaSelectElement (doc, child);
    }
  return colhead;
}

/*----------------------------------------------------------------------
  RemoveColumn remove the current colhead if it's empty.
  The parameter ifEmpty makes removing optional.
  Returns TRUE if the column has been removed.
  ----------------------------------------------------------------------*/
ThotBool RemoveColumn (Element colhead, Document doc, ThotBool ifEmpty,
		       ThotBool inMath)
{
  Element             row, firstrow;
  Element             cell, group, table;
  ElementType         elType;
  AttributeType       attrType;
  Attribute           attr;
  int                 rowType, colspan;
  ThotBool            empty, span;

  empty = FALSE; /* return TRUE if the column is deleted */
  if (colhead == NULL)
    return (empty);

  /* check if the column becomes empty */
  elType = TtaGetElementType (colhead);
  attrType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    {
      elType.ElTypeNum = MathML_EL_MTABLE;
      rowType = MathML_EL_TableRow;
      attrType.AttrTypeNum = MathML_ATTR_columnspan;
    }
  else
    {
      elType.ElTypeNum = HTML_EL_Table;
      rowType = HTML_EL_Table_row;
      attrType.AttrTypeNum = HTML_ATTR_colspan_;
    }

  table = TtaGetTypedAncestor (colhead, elType);
  elType.ElTypeNum = rowType;
  firstrow = TtaSearchTypedElement (elType, SearchInTree, table);
  if (colhead != NULL && firstrow != NULL)
    {
      empty = TRUE;  /* TRUE when all cells are empty */
      if (ifEmpty)
	{
	  row = firstrow;
	  /* the rows group could be thead, tbody, tfoot */
	  group = TtaGetParent (row);
	  while (row && empty)
	    {
	      cell = GetCellFromColumnHead (row, colhead, inMath);
	      if (cell != NULL && TtaGetElementVolume (cell) != 0)
		empty = FALSE;
	      else
		{
		row = GetSiblingRow (row, FALSE, inMath);
	        /* do we have to get a new group of rows */
	        if (row == NULL && !inMath)
		  {
		  elType = TtaGetElementType (group);
		  if (elType.ElTypeNum == HTML_EL_tbody)
		    {
		      row = group;
		      group = TtaGetParent (group);
		      TtaNextSibling (&row);
		      if (row != NULL)
			group = row;
		      else
			TtaNextSibling (&group);
		    }
		  else
		    {
		      TtaNextSibling (&group);
		      if (group != NULL)
			{
			  elType = TtaGetElementType (group);
			  if (elType.ElTypeNum == HTML_EL_Table_body)
			    group = TtaGetFirstChild (group);
			}
		    }
		  if (group != NULL)
		    row = TtaGetFirstChild (group);
		  else
		    row = NULL;
		  }
		}
	    }
	}

      if (empty)
	{
	  span = FALSE; /* TRUE when a cell with a span value is deleted */
	  /* remove the current column */
	  row = firstrow;
	  /* the rows group could be thead, tbody, or tfoot */
	  group = TtaGetParent (row);
	  while (row != NULL)
	    {
	      cell = GetCellFromColumnHead (row, colhead, inMath);
	      if (cell != NULL)
		{
		  if (!inMath)
		    {
		      /* look at whether it's a spanning column */
		      attr = TtaGetAttribute (cell, attrType);
		      if (attr)
			{
			  colspan = TtaGetAttributeValue (attr);
			  if (colspan == 0 || colspan > 1)
			    /* colspan = 0 means infinite span */
			    span = TRUE;
			}
		    }
		  /* register the deleted cells only if the column is
		     deleted explicitely by the user */
		  if (!ifEmpty)
		    if (TtaPrepareUndo (doc))
		      {
		        TtaRegisterElementDelete (cell, doc);
                        /* change the value of "info" in the latest cell
			   deletion recorded in the Undo queue. The goal is to
			   allow procedure CellPasted to regenerate only one
			   column head when undoing the operation */
			TtaChangeInfoLastRegisteredElem (doc, 3);
		      }
		  TtaDeleteTree (cell, doc);
		}
	      row = GetSiblingRow (row, FALSE, inMath);
	      /* do we have to get a new group of rows */
	      if (row == NULL && !inMath)
		{
		  elType = TtaGetElementType (group);
		  if (elType.ElTypeNum == HTML_EL_tbody)
		    {
		      row = group;
		      group = TtaGetParent (group);
		      TtaNextSibling (&row);
		      if (row != NULL)
			group = row;
		      else
			TtaNextSibling (&group);
		    }
		  else
		    {
		      TtaNextSibling (&group);
		      if (group != NULL)
			{
			  elType = TtaGetElementType (group);
			  if (elType.ElTypeNum == HTML_EL_Table_body)
			    group = TtaGetFirstChild (group);
			}
		    }
		  if (group != NULL)
		    row = TtaGetFirstChild (group);
		  else
		    row = NULL;
		}
	    }
	  TtaDeleteTree (colhead, doc);
	  if (!ifEmpty)
	    if (TtaPrepareUndo (doc))
	      /* The value of "info" in the latest cell deletion recorded in
		 the Undo queue should be 4 to allow procedure CellPasted to
		 regenerate a column head for the last cell when undoing
		 the operation */
		TtaChangeInfoLastRegisteredElem (doc, 4);
	  if (span)
	     CheckAllRows (table, doc, FALSE, FALSE);
	}
    }
  return (empty);
}

#define MAX_COLS 100
/*----------------------------------------------------------------------
  CheckAllRows
  ----------------------------------------------------------------------*/
void CheckAllRows (Element table, Document doc, ThotBool placeholder,
		   ThotBool deleteLastEmptyColumns)
{
  Element            *colElement, *delayedColExt;
  Element             row, nextRow, firstrow, colhead, prevColhead;
  Element             cell, nextCell, group, prevGroup, new_, prev;
  ElementType         elType;
  AttributeType       attrTypeHSpan, attrTypeVSpan, attrType;
  Attribute           attr;
  SSchema             tableSS;
  int                *colVSpan;
  int                 span, cRef, cNumber, cDelayed, overflow;
  int                 i, rowType;
  ThotBool            inMath;

  NewTable = FALSE;
  if (table == NULL)
    return;
  colElement = (Element*)TtaGetMemory (sizeof (Element) * MAX_COLS);
  delayedColExt = (Element*)TtaGetMemory (sizeof (Element) * MAX_COLS);
  colVSpan = (int *)TtaGetMemory (sizeof (int) * MAX_COLS);
  /* store the list of colheads */
  elType = TtaGetElementType (table);
  tableSS = elType.ElSSchema;
  attrType.AttrSSchema = tableSS;
  attrTypeHSpan.AttrSSchema = tableSS;
  attrTypeVSpan.AttrSSchema = tableSS;
  inMath = TtaSameSSchemas (tableSS, TtaGetSSchema ("MathML", doc));
  if (inMath)
    {
      elType.ElTypeNum = MathML_EL_MColumn_head;
      rowType = MathML_EL_TableRow;
      attrTypeHSpan.AttrTypeNum = MathML_ATTR_columnspan;
      attrTypeVSpan.AttrTypeNum = MathML_ATTR_rowspan_;
      attrType.AttrTypeNum = MathML_ATTR_MColExt;
    }
  else
    {
      elType.ElTypeNum = HTML_EL_Column_head;
      rowType = HTML_EL_Table_row;
      attrTypeHSpan.AttrTypeNum = HTML_ATTR_colspan_;
      attrTypeVSpan.AttrTypeNum = HTML_ATTR_rowspan_;
      attrType.AttrTypeNum = HTML_ATTR_ColExt;
    }

  colhead = TtaSearchTypedElement (elType, SearchInTree, table);
  cNumber = 0;
  cDelayed = 0;
  while (colhead != 0 && cNumber < MAX_COLS)
    {
      colElement[cNumber] = colhead;
      colVSpan[cNumber] = 0;
      TtaNextSibling (&colhead);
      cNumber++;
    }

  cell = NULL;
  elType.ElTypeNum = rowType;
  firstrow = TtaSearchTypedElement (elType, SearchInTree, table);
  if (cNumber != 0 && firstrow != NULL)
    {
    row = firstrow;
    /* the rows group could be thead, tbody, tfoot */
    group = TtaGetParent (row);
    while (row)
      {
      nextRow = row;
      TtaNextSibling (&nextRow);
      elType = TtaGetElementType (row);
      if ((!inMath && elType.ElTypeNum == rowType) ||
	  (inMath && (elType.ElTypeNum == MathML_EL_MTR ||
		      elType.ElTypeNum == MathML_EL_MLABELEDTR)))
	{
	/* treat all cells in the row */
	cRef = 0;
	cell = NULL;
	/* get the first cell in the row, ignoring other elements such as
	   comments or elements from other namespaces */
	nextCell = TtaGetFirstChild (row);
	if (nextCell)
	  elType = TtaGetElementType (nextCell);
	/* skip non-cell elements */
	while (nextCell &&
	       (elType.ElSSchema != tableSS ||
		(inMath && elType.ElTypeNum != MathML_EL_MTD) ||
		(!inMath && (elType.ElTypeNum != HTML_EL_Data_cell &&
			     elType.ElTypeNum != HTML_EL_Heading_cell))))
	  {
	    TtaNextSibling (&nextCell);
	    if (nextCell)
	      elType = TtaGetElementType (nextCell);
	  }
	if (nextCell && inMath &&
	    TtaGetElementType(row).ElTypeNum == MathML_EL_MLABELEDTR)
	  /* skip the first significant child of the row: it's a label */
	  {
	  /* skip comments and PIs first */
	  do
	    {
	    elType = TtaGetElementType (nextCell);
	    if (elType.ElTypeNum == MathML_EL_XMLcomment ||
		elType.ElTypeNum == MathML_EL_XMLPI)
	      TtaNextSibling (&nextCell);
	    }
	  while (nextCell &&
		 (elType.ElTypeNum == MathML_EL_XMLcomment ||
		  elType.ElTypeNum == MathML_EL_XMLPI));
	  /* skip the following element */
	  nextCell = GetSiblingCell (nextCell, FALSE, inMath);
	  }
	while (nextCell)
	  {
	  cell = nextCell;
          TtaNextSibling (&nextCell);
	  elType = TtaGetElementType (cell);
	  if (!inMath && elType.ElTypeNum == HTML_EL_Table_cell)
	    {
	    /* replace a Table_cell by a Data_cell */
	    elType.ElTypeNum = HTML_EL_Data_cell;
	    new_ = TtaNewTree (doc, elType, "");
	    TtaInsertFirstChild (&new_, cell, doc);
	    cell = new_;
	    }
	  
	  /* process only cell elements */
	  if ((inMath && elType.ElTypeNum == MathML_EL_MTD) ||
	      (!inMath && (elType.ElTypeNum == HTML_EL_Data_cell ||
			   elType.ElTypeNum == HTML_EL_Heading_cell)))
	    {
	    /* is there any rowspan attribute in previous row */
	    while (cRef < cNumber && colVSpan[cRef] > 0)
	      {
		if (colVSpan[cRef] < THOT_MAXINT)
		  colVSpan[cRef]--;
		cRef++;
	      }
	    
	    if (cRef == cNumber && cRef < MAX_COLS)
	      {
	      /* there is no Column_head for that cell */
	      /* Create an additional Column_head */
	      colElement[cRef] = NewColumnHead (colElement[cRef - 1], FALSE,
						TRUE, row, doc, inMath, TRUE);
	      colVSpan[cRef] = 0;
	      cNumber++;
	      }
	    
	    if (cRef < cNumber)
	      {
	      /* link the cell to its corresponding Column_head */
	      LinkCellToColumnHead (cell, colElement[cRef], doc, inMath);

	      /* is there an attribute rowspan for that cell ? */
	      attr = TtaGetAttribute (cell, attrTypeVSpan);
	      if (attr == NULL)
		span = 1;
	      else
		{
		  span = TtaGetAttributeValue (attr);
		  /* Set the attribute RowExt if row span > 1 */
		  if (span > 1)
		    colVSpan[cRef] = span - 1;
		  else if (span == 0)
		    colVSpan[cRef] = THOT_MAXINT;
		}
	      overflow = SetRowExt (cell, span, doc, inMath);
	      if (overflow > 0 && attr)
		/* spanning value is too great. update it */
		TtaSetAttributeValue (attr, span-overflow, cell, doc);

	      /* if there an attribute colspan for that cell,
		 update attribute ColExt */
	      attr = TtaGetAttribute (cell, attrTypeHSpan);
	      if (attr == NULL)
		/* no colspan attribute */
		span = 1;
	      else
		{
		span = TtaGetAttributeValue (attr);
		if (span == 0)
		  /* infinite span */
		  span = THOT_MAXINT;
		if (span > 1)
		  {
		  i = 1;
		  while (i < span && cRef + 1 < cNumber)
		    {
		      cRef++;
		      colVSpan[cRef] = colVSpan[cRef-1];
		      i++;
		    }
		  if (GetSiblingCell (cell, FALSE, inMath))
		    /* There are more cells in this row, after the spanning
		       cell. Create additional Column_heads for the spanning
		       cell, except if spanning is "infinite". */
		    {
		    if (span < THOT_MAXINT)
		      while (i < span)
			{
			if (cRef + 1 < MAX_COLS)
			  {
			  cRef++;
			  colElement[cRef] = NewColumnHead (colElement[cRef-1],
					  FALSE, TRUE, row, doc, inMath, TRUE);
			  cNumber++;
			  }
			colVSpan[cRef] = colVSpan[cRef-1];
			i++;
			}
		    }
		  else if (i < span)
		    {
		      /* The colspan cannot be applied now */
		      delayedColExt[cDelayed++] = cell;
		      span = 1;
		    }
		  
		  }
		}
	      SetColExt (cell, span, doc, inMath, FALSE);

	      cRef++;
	      }
	    }

	  /* accept comments, PIs and invalid elements */
	  else if ((!inMath &&
		    elType.ElTypeNum != HTML_EL_Invalid_element &&
		    elType.ElTypeNum != HTML_EL_Comment_ &&
		    elType.ElTypeNum != HTML_EL_XMLPI) ||
		   (inMath &&
		    elType.ElTypeNum != MathML_EL_XMLcomment &&
		    elType.ElTypeNum != MathML_EL_XMLPI))
	    /* Delete any other type of element */
	    TtaDeleteTree (cell, doc);
	  }
	
	/* check missing cells */
	while (cRef < cNumber)
	  {
	  if (colVSpan[cRef] > 0)
	    {
	      if (colVSpan[cRef] < THOT_MAXINT)
		colVSpan[cRef]--;
	    }
	  else
	    cell = AddEmptyCellInRow (row, colElement[cRef], cell, FALSE, doc,
				      inMath, placeholder, TRUE);
	  cRef++;
	  }
	}
      /* accept comments, PIs and invalid elements */
      else if ((!inMath &&
		elType.ElTypeNum != HTML_EL_Invalid_element &&
		elType.ElTypeNum != HTML_EL_Comment_ &&
		elType.ElTypeNum != HTML_EL_XMLPI) ||
	       (inMath &&
		elType.ElTypeNum != MathML_EL_XMLcomment &&
		elType.ElTypeNum != MathML_EL_XMLPI))
	/* Delete any other type of element */
	TtaDeleteTree (row, doc);
      row = nextRow;

      /* do we have to get a new group of rows */
      if (row == NULL && !inMath)
	{
	elType = TtaGetElementType (group);
	if (elType.ElTypeNum == HTML_EL_tbody)
	  /* this is a tbody. Is there a sibling tboby in the same
	     Table_body parent? */
	  {
	  row = group;
	  group = TtaGetParent (group); /* the Table_body parent */
	  TtaNextSibling (&row); /* next sibling of the current tbody */
	  if (row != NULL)
	    {
	    /* skip comments and other insignificant elements, until we get
	       a sibling tbody */
	    elType = TtaGetElementType (row);
	    while (row && (elType.ElSSchema != tableSS ||
			   elType.ElTypeNum != HTML_EL_tbody))
	      {
		prev = row;
		TtaNextSibling (&row);
		/* remove text leaves between tbody elements */
		if (elType.ElTypeNum == HTML_EL_TEXT_UNIT)
		  TtaDeleteTree (prev, doc);
		if (row)
		  elType = TtaGetElementType (row);
	      }
	    if (row)
	      group = row;
	    }
	  if (row == NULL)
	    /* there is no sibling tbody. Get the next sibling of the
	       Table_body element */
	    TtaNextSibling (&group);
	  }
	else
	  {
	  TtaNextSibling (&group);
	  if (group != NULL)
	    {
	    elType = TtaGetElementType (group);
	    if (elType.ElTypeNum == HTML_EL_Table_body)
	      group = TtaGetFirstChild (group);
	    }
	  }
	if (group != NULL)
	  {
	  elType = TtaGetElementType (group);
	  /* skip comments, PIs and invalid elements */
	  while (group &&
		 (elType.ElTypeNum == HTML_EL_TEXT_UNIT ||
		  elType.ElTypeNum == HTML_EL_Invalid_element ||
		  elType.ElTypeNum == HTML_EL_Comment_ ||
		  elType.ElTypeNum == HTML_EL_XMLPI))
	    {
	      prevGroup = group;
	      TtaNextSibling (&group);
	      if (elType.ElTypeNum == HTML_EL_TEXT_UNIT)
		TtaDeleteTree (prevGroup, doc);
	      if (group)
		elType = TtaGetElementType (group);
	    }
	  if (group)
	    {
	      elType = TtaGetElementType (group);
	      if (elType.ElTypeNum == HTML_EL_Table_foot)
		/* don't look for rows in the Table_foot! */
		row = NULL;
	      else
		row = TtaGetFirstChild (group);
	    }
	  else
	    row = NULL;
	  }
	else
	  /* no more group */
	  row = NULL;
	if (row)
	  /* we are starting with a new group of rows. Ignore vertical 
	     spanning from previous group */
	  for (i = 0; i < cNumber; i++)
	    colVSpan[i] = 0;
	}
      }
    }

  /* if there are some empty columns at the end, remove them */
  if (deleteLastEmptyColumns && cNumber > 0)
    {
    /* start with the last column of the table */
    colhead = colElement[cNumber - 1];
    while (colhead)
      {
      /* get the previous column head before it is deleted */
      prevColhead = colhead;
      TtaPreviousSibling (&prevColhead);
      /* remove this column if it is empty */
      if (!RemoveColumn (colhead, doc, TRUE, inMath))
	/* this column is not empty and has not been removed. Stop */
	colhead = NULL;
      else
	/* the column was empty and has been removed */
	colhead = prevColhead;     
      }
    }

  /* apply delayed colspan */
  if (cDelayed > 0)
    {
      i = 0;
      while (i < cDelayed)
	{
	  attr = TtaGetAttribute (delayedColExt[i], attrTypeHSpan);
	  if (attr)
	    {
	      span = TtaGetAttributeValue (attr);
	      ChangeColspan (delayedColExt[i], 1, span, doc);
	      overflow = SetColExt (delayedColExt[i], span, doc, inMath,FALSE);
	      if (overflow > 0)
		/* spanning value is too great. update it */
		TtaSetAttributeValue (attr, span-overflow, delayedColExt[i],
				      doc);
	    }
	  i++;
	}
    }

  TtaFreeMemory (colElement);
  TtaFreeMemory (delayedColExt);
  TtaFreeMemory (colVSpan);
}

/*----------------------------------------------------------------------
   CheckTable      Check a table and create the missing elements.  
  ----------------------------------------------------------------------*/
void CheckTable (Element table, Document doc)
{
  ElementType         elType;
  Element             el, columnHeads, thead, tfoot, firstcolhead,
                      tbody, Tablebody, firstgroup, prevrow,
                      foot, prevEl, nextEl,
		      enclosingTable;
  AttributeType       attrType;
  Attribute           attr;
  int                 PreviousStuctureChecking;
  ThotBool            before;

  firstcolhead = NULL;
  PreviousStuctureChecking = 0;
  if (table != NULL)
    {
      /* what are the children of element table? */
      columnHeads = NULL;
      thead = NULL;
      tfoot = NULL;
      Tablebody = NULL;
      el = TtaGetFirstChild (table);
      while (el != NULL)
	{
	  elType = TtaGetElementType (el);
	  if (elType.ElTypeNum == HTML_EL_Table_head)
	    columnHeads = el;
	  else if (elType.ElTypeNum == HTML_EL_Table_body)
	    {
	      if (Tablebody == NULL)
		Tablebody = el;
	    }
	  else if (elType.ElTypeNum == HTML_EL_thead)
	    thead = el;
	  else if (elType.ElTypeNum == HTML_EL_tfoot)
	    tfoot = el;
	  TtaNextSibling (&el);
	}
      if (columnHeads != NULL)
	/* this table has already been checked */
	return;

      /* if this table is within another table and if it has no border
	 attribute, create an attribute border=0 to avoid inheritance
	 of the the border attribute from the enclosing table */
      elType = TtaGetElementType (table);
      enclosingTable = TtaGetTypedAncestor (table, elType);
      if (enclosingTable != NULL)
	/* there is an enclosing table */
	{
	  attrType.AttrSSchema = elType.ElSSchema;
	  attrType.AttrTypeNum = HTML_ATTR_Border;
	  if (TtaGetAttribute (enclosingTable, attrType) != NULL)
	    /* the enclosing table has a Border attribute */
	    if (TtaGetAttribute (table, attrType) == NULL)
	      /* the current table has no Border attribute */
	      {
		attr = TtaNewAttribute (attrType);
		TtaAttachAttribute (table, attr, doc);
		TtaSetAttributeValue (attr, 0, table, doc);
	      }
	}

      /* disable document structure checking */
      PreviousStuctureChecking = TtaGetStructureChecking (doc);
      TtaSetStructureChecking (0, doc);

      /* create a Table_head element with a first Column_head */
      elType = TtaGetElementType (table);
      elType.ElTypeNum = HTML_EL_Table_head;
      columnHeads = TtaNewTree (doc, elType, "");
      if (columnHeads != NULL)
	{
	  firstcolhead = TtaGetFirstChild (columnHeads);
	  el = TtaGetFirstChild (table);
	  if (el == NULL)
	    TtaInsertFirstChild (&columnHeads, table, doc);
	  else
	    {
	      /* skip elements Comment and Invalid_element and insert the new
	         element Table_head, after the element caption if it is
		 present or as the first child if there is no caption */
	      while (el != NULL)
		{
		  elType = TtaGetElementType (el);
		  if (elType.ElTypeNum == HTML_EL_Invalid_element ||
		      elType.ElTypeNum == HTML_EL_Comment_ ||
		      elType.ElTypeNum == HTML_EL_XMLPI)
		    TtaNextSibling (&el);
		  else
		    {
		      before = elType.ElTypeNum != HTML_EL_CAPTION;
		      TtaInsertSibling (columnHeads, el, before, doc);
		      el = NULL;
		    }
		}
	    }

	  if (Tablebody == NULL)
	    {
	      /* There is no Table_body element. Create a Table_body element */
	      elType.ElTypeNum = HTML_EL_Table_body;
	      Tablebody = TtaNewElement (doc, elType);
	      if (Tablebody != NULL)
		{
		  if (thead != NULL)
		    TtaInsertSibling (Tablebody, thead, FALSE, doc);
		  else
		    TtaInsertSibling (Tablebody, columnHeads, FALSE, doc);
		  /* collect all Table_row elements and put them in the new
		     Table_body */
		  tbody = NULL;
		  el = Tablebody;
		  TtaNextSibling (&el);
		  prevrow = NULL;
		  prevEl = NULL;
		  attrType.AttrSSchema = elType.ElSSchema;
		  while (el != NULL)
		    {
		      elType = TtaGetElementType (el);
		      nextEl = el;
		      TtaNextSibling (&nextEl);
		      if (elType.ElTypeNum == HTML_EL_Table_row)
			{
			  TtaRemoveTree (el, doc);
			  if (prevrow == NULL)
			    {
			      /* create a tbody element */
			      elType.ElTypeNum = HTML_EL_tbody;
			      tbody = TtaNewElement (doc, elType);
			      if (prevEl == NULL)
				TtaInsertFirstChild (&tbody, Tablebody, doc);
			      else
				TtaInsertSibling (tbody, prevEl, FALSE, doc);
			      TtaInsertFirstChild (&el, tbody, doc);
			    }
			  else
			    TtaInsertSibling (el, prevEl, FALSE, doc);
			  prevrow = el;
			  prevEl = el;
			}
		      else
			{
			  if (elType.ElTypeNum == HTML_EL_tbody)
			    {
			      if (prevrow != NULL)
				prevEl = TtaGetParent (prevrow);
			      prevrow = NULL;
			    }
			  TtaRemoveTree (el, doc);
			  if (prevEl == NULL)
			    TtaInsertFirstChild (&el, Tablebody, doc);
			  else
			    TtaInsertSibling (el, prevEl, FALSE, doc);
			  prevEl = el;
			}
		      el = nextEl;
		    }
		}
	    }

	  /* create a Table_foot element at the end */
	  elType.ElTypeNum = HTML_EL_Table_foot;
	  foot = TtaNewTree (doc, elType, "");
	  if (foot != NULL)
	    {
	    if (tfoot != NULL)
	      {
		/* move element tfoot at the end */
		TtaRemoveTree (tfoot, doc);
		TtaInsertSibling (tfoot, Tablebody, FALSE, doc);
		TtaInsertSibling (foot, tfoot, FALSE, doc);
	      }
	    else
	      TtaInsertSibling (foot, Tablebody, FALSE, doc);
	    }
	  
	  if (thead != NULL)
	    firstgroup = thead;
	  else if (Tablebody != NULL)
	    firstgroup = Tablebody;
	  else
	    firstgroup = tfoot;

	  /* associate each cell with a column */
	  CheckAllRows (table, doc, FALSE, FALSE);
	}
    }
  /* resume document structure checking */
  TtaSetStructureChecking ((ThotBool)PreviousStuctureChecking, doc);
}

/*----------------------------------------------------------------------
   SpanningCellForRow
   returns the cell in the column represented by colhead that spans the row
   element. Return NULL if there is no spanning cell in this column for
   that row.
   If a spanning cell is found, parameter colspan returns its colspan value.
   If update is TRUE, the spanning of the cell is increased (if we are adding
   a new row) or decreased (when removing a row).
  ----------------------------------------------------------------------*/
static Element SpanningCellForRow (Element row, Element colhead,
				   Document doc, ThotBool inMath,
				   ThotBool update, ThotBool addRow,
				   int *colspan)
{
  Element             cell, currentRow, lastRow;
  ElementType         elType;
  Attribute           attr;
  AttributeType       attrType;
  int                 rowspan, pos;
  ThotBool            lastRemoved, newLastRow;

  *colspan = 1;
  if (row == NULL)
    return NULL;
  elType = TtaGetElementType (row);

  /* look at the rows before the row of interest and find the first row
     that contains a cell for the column of interest */
  pos = 0;
  cell = NULL;
  currentRow = row;
  do
    {
      currentRow = GetSiblingRow (currentRow, TRUE, inMath);
      if (currentRow)
	{
	  cell = GetCellFromColumnHead (currentRow, colhead, inMath);
	  pos++;
	}
    }
  while (!cell && currentRow);

  if (cell)
    /* we have found a cell in a row above (currentRow) */
    {
      /* check the rowspan attribute of the cell */
      attrType.AttrSSchema = elType.ElSSchema;
      if (inMath)
	attrType.AttrTypeNum = MathML_ATTR_rowspan_;
      else
	attrType.AttrTypeNum = HTML_ATTR_rowspan_;
      attr = TtaGetAttribute (cell, attrType);
      if (attr)
	/* this cell has a rowspan attribute */
	{
	  rowspan = TtaGetAttributeValue (attr);
	  if (rowspan == 0 || rowspan - pos > 0)
	    /* this cell spans the row of interest */
	    /* rowspan = 0 means that the cell spans to the end of the
	       current rows group */
	    {
	    if (update)
              {
	      lastRemoved = FALSE;
              newLastRow = FALSE;
	      if (rowspan == 0)
		{
		  if (!GetSiblingRow (row, FALSE, inMath))
		    /* the cell is spanning to the last row and we are
		       changing the last row */
		    {
		      if (addRow)
			newLastRow = TRUE;
		      else
			lastRemoved = TRUE;
		    }
		} 
	      else
		/* update the spanning */
		if (addRow)
		  rowspan++;
		else
		  {
		    rowspan--;
		    if (rowspan - pos == 0)
		      /* we are removing the last row spanned by the cell */
		      lastRemoved = TRUE;
		  }
	      if (rowspan == 1)
	        /* no spanning any more */
		{
		  /* do not register this deletion of attribute rowspan in the
		     undo queue: undo, when undoing a row creation, would send
		     the event for the attribute before the row is actually
		     removed. RowDeleted called by undo will do the job */
		  if (!addRow)
		    TtaRegisterAttributeDelete (attr, cell, doc);
		  TtaRemoveAttribute (cell, attr, doc);
		  if (inMath)
		    attrType.AttrTypeNum = MathML_ATTR_MRowExt;
		  else
		    attrType.AttrTypeNum = HTML_ATTR_RowExt;
		  attr = TtaGetAttribute (cell, attrType);
		  if (attr)
		    TtaRemoveAttribute (cell, attr, doc);
		  lastRemoved = FALSE;
		}
	      else if (rowspan != 0)
		{
		  /* do not register this change of attribute rowspan in the
		     undo queue: undo, when undoing a row creation, would send
		     the event for the attribute before the row is actually
		     removed. RowDeleted called by undo will do the job */
		  if (!addRow)
		    TtaRegisterAttributeReplace (attr, cell, doc);
		  TtaSetAttributeValue (attr, rowspan, cell, doc);
		}
	      lastRow = NULL;
	      if (newLastRow)
		lastRow = row;
	      if (lastRemoved)
		lastRow = GetSiblingRow (row, TRUE, inMath);
	      if (lastRow)
		/* update the RowExt attribute */
		{
		  if (inMath)
		    attrType.AttrTypeNum = MathML_ATTR_MRowExt;
		  else
		    attrType.AttrTypeNum = HTML_ATTR_RowExt;
		  attr = TtaGetAttribute (cell, attrType);
		  if (attr)
		    {
		      if (currentRow == lastRow)
			{
			  TtaRemoveAttribute (cell, attr, doc);
			  attr = NULL;
			}
		    }
		  else if (currentRow != lastRow)
		    {
		      attr = TtaNewAttribute (attrType);
		      if (attr != NULL)
			TtaAttachAttribute (cell, attr, doc);
		    }
		  if (attr)
		    TtaSetAttributeReference (attr, cell, doc, lastRow, doc);
		}
	      /* we have finished updating */
	      }
	    /* get the colspan value of the rowspanning cell */
	    if (inMath)
	      attrType.AttrTypeNum = MathML_ATTR_columnspan;
	    else
	      attrType.AttrTypeNum = HTML_ATTR_colspan_;
	    attr = TtaGetAttribute (cell, attrType);
	    if (attr)
	      /* this cell has an attribute colspan */
	      *colspan = TtaGetAttributeValue (attr);
	    if (*colspan < 0)
	      *colspan = 1;
	    }
	  else
	    cell = NULL;
	}
      else
	cell = NULL;
    }
  return (cell);
}

/*----------------------------------------------------------------------
  NewCell  a new cell has been created or pasted in a HTML table.
  Link it to its column head.
  If generateColumn is TRUE, the new cell generates a new column head.
  If generateEmptyCells is TRUE, generate empty cells for the same column
  in other rows.
  ----------------------------------------------------------------------*/
void NewCell (Element cell, Document doc, ThotBool generateColumn,
	      ThotBool generateEmptyCells)
{
  Element             newcell, row, colhead, lastColhead, chead, pcell;
  ElementType         elType;
  AttributeType       attrTypeCspan, attrTypeRefC;
  Attribute           attr;
  DisplayMode         dispMode;
  char                ptr[100];
  int                 span, i, col;
  ThotBool            before, inMath;

  dispMode = TtaGetDisplayMode (doc);
  if (dispMode == DisplayImmediately)
    TtaSetDisplayMode (doc, DeferredDisplay);

  elType = TtaGetElementType (cell);
  attrTypeCspan.AttrSSchema = elType.ElSSchema;
  attrTypeRefC.AttrSSchema = elType.ElSSchema;
  inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
  if (!inMath && elType.ElTypeNum == HTML_EL_Table_cell)
    /* change the cell into a Data_cell */
    {
      elType.ElTypeNum = HTML_EL_Data_cell;
      newcell = TtaNewTree (doc, elType, "");
      TtaInsertFirstChild (&newcell, cell, doc);
      cell = newcell;
    }
  else
    newcell = cell;
  if (inMath)
    {
      elType.ElTypeNum = MathML_EL_TableRow;
      attrTypeRefC.AttrTypeNum = MathML_ATTR_MRef_column;
      attrTypeCspan.AttrTypeNum = MathML_ATTR_columnspan;
    }
  else
    {
      elType.ElTypeNum = HTML_EL_Table_row;
      attrTypeRefC.AttrTypeNum = HTML_ATTR_Ref_column;
      attrTypeCspan.AttrTypeNum = HTML_ATTR_colspan_;
    }
  /* get the enclosing row element */
  row = TtaGetTypedAncestor (cell, elType);
  /* locate the previous or the next column head */
  colhead = NULL;
  cell = GetSiblingCell (newcell, TRUE, inMath);
  if (cell == NULL)
    /* no cell before the new cell. Try after the new cell */
    {
      cell = GetSiblingCell (newcell, FALSE, inMath);
      before = TRUE;
      if (cell)
	/* there is a following cell. Get its column head */
	{
	  attr = TtaGetAttribute (cell, attrTypeRefC);
	  if (attr)
	    {
	      TtaGiveReferenceAttributeValue (attr, &colhead, ptr, &i);
	      if (colhead)
		/* the column head is found */
		{
		  /* if there are some cells spanning from above rows,
		     skip them. Check previous column heads. */
		  chead = colhead;
		  TtaPreviousSibling (&chead);
		  col = 1;
		  while (chead)
		    {
		      pcell = SpanningCellForRow (row, chead, doc, inMath,
						  FALSE, FALSE, &span);
		      if (pcell)
			/* there is a cell linked to that column head that
			   spans our row */
			{
			  if (span == col)
			    /* its colspan make it touch our cell */
			    colhead = chead;
			  else if (span < col)
			    /* it leaves room for the new cell. stop */
			    chead = NULL;
			}
		      if (chead)
			{
			  TtaPreviousSibling (&chead);
			  col++;
			}
		    }

		  attr = TtaGetAttribute (newcell, attrTypeCspan);
		  if (attr)
		    /* the new cell has an attribute colspan */
		    {
		      span = TtaGetAttributeValue (attr);
		      /* ignore span if zero: there is a cell after this
			 newcell with an "infinite" span. Strange */
		      if (span > 1)
			/* skip the previous column heads according to the
			   span of the new cell */
			{
			  for (i = 1; i < span && colhead; i++)
			    {
			      lastColhead = colhead;
			      TtaPreviousSibling (&colhead);
			    }
			  if (!colhead)
			    colhead = lastColhead;
			}
		    }
		}
	    }
	}
    }
  else
    /* there is a cell before the new cell */
    {
      before = FALSE;
      /* get its column head */
      attr = TtaGetAttribute (cell, attrTypeRefC);
      if (attr != NULL)
	{
	  TtaGiveReferenceAttributeValue (attr, &colhead, ptr, &i);
	  if (colhead)
	    {
	      /*if this cell has an attribute colspan, skip some column heads*/
	      attr = TtaGetAttribute (cell, attrTypeCspan);
	      if (attr)
		{
		  span = TtaGetAttributeValue (attr);
		  /* ignore colspan if zero: the new cell would be after this
		     supposed infinite cell. Strange */
		  if (span > 1)
		    {
		      for (i = 1; i < span && colhead; i++)
			{
			  lastColhead = colhead;
			  TtaNextSibling (&colhead);
			}
		      if (!colhead)
			colhead = lastColhead;
		    }
		}
	      /* if it's the creation of a new column, create the new
                 column here. Otherwise take the existing cells into
	         account */
	      if (!generateColumn)
	        /* if there are some cells spanning from above rows,
		   skip them */
	        do
		  {
		  chead = colhead;
		  TtaNextSibling (&chead);
		  if (chead)
		    pcell = SpanningCellForRow (row, chead, doc, inMath, FALSE,
						FALSE, &span);
		  else
		    pcell = NULL;
		  if (pcell)
		    {
		      colhead = chead;
		      /* if this cell has an attribute colspan, skip some
			 column heads */
		      attr = TtaGetAttribute (pcell, attrTypeCspan);
		      if (attr)
			{
			  span = TtaGetAttributeValue (attr);
			  if (span > 1)
			    {
			      for (i = 1; i < span && colhead; i++)
				{
				  lastColhead = colhead;
				  TtaNextSibling (&colhead);
				}
			      if (!colhead)
				colhead = lastColhead;
			    }
			}
		    }
		  }
	        while (pcell);
	    }
	}
    }

  if (colhead)
    {
      if (generateColumn)
	/* generate the new column */
	colhead = NewColumnHead (colhead, before, FALSE, row, doc, inMath,
				 generateEmptyCells);
      else if (before)
	/* select the previous column */
	TtaPreviousSibling (&colhead);
      else
	/* select the next column */
	TtaNextSibling (&colhead);
      /* next pasted cells with info = 3 should be linked to this column */
      CurrentColumn = colhead;
      /* link the new cell to the new colhead */
      LinkCellToColumnHead (newcell, colhead, doc, inMath);
    }
   TtaSetDisplayMode (doc, dispMode);
}

/*----------------------------------------------------------------------
 CreateCell
 The user wants to create a new cell (i.e. a new column). If the new cell
 is supposed to be right a cell that has attribute colspan = 0,
 the creation id refused.
 -----------------------------------------------------------------------*/
ThotBool CreateCell (NotifyElement *event)
{
  Element        row, cell;
  ElementType    elType;
  Attribute      attr;
  AttributeType  attrType;
  int            i;
  ThotBool       forbidden;

  forbidden = FALSE;
  row = event->element;
  if (event->position > 0)
    {
      cell = TtaGetFirstChild (row);
      for (i = 1; i < event->position && cell; i++)
	TtaNextSibling (&cell);
      if (cell)
	{
	  elType = TtaGetElementType(row);
	  attrType.AttrSSchema = elType.ElSSchema;
          if (TtaSameSSchemas (elType.ElSSchema,
			       TtaGetSSchema ("MathML", event->document)))
	    attrType.AttrTypeNum = MathML_ATTR_columnspan;
	  else
	    attrType.AttrTypeNum = HTML_ATTR_colspan_;
	  attr = TtaGetAttribute (cell, attrType);
	  if (attr)
	    if (TtaGetAttributeValue (attr) == 0)
	      forbidden = TRUE;
	}
    }
  return forbidden;
}

/*----------------------------------------------------------------------
   CellCreated
   a new cell has been created in a table
  ----------------------------------------------------------------------*/
void CellCreated (NotifyElement * event)
{
  Element             cell, row;
  Document            doc;

  if (event->info == 1)
    /* the creation is already done by undo */
    return;
  cell = event->element;
  doc = event->document;
  row = TtaGetParent (cell);
  if (row == CurrentRow)
    {
      /* the new cell belongs to the newly created row. Already processed */
      /* by RowCreated. */
      if (cell == CurrentCell)
	CurrentRow = NULL;
    }
  else
    /* a new cell in an existing row */
    {
      /* the creation of the cell must be registered after
	 the created colhead */
      TtaCancelLastRegisteredOperation (doc);
      NewCell (cell, doc, TRUE, TRUE);
      TtaRegisterElementCreate (cell, doc);
      /* change the value of "info" in the latest cell
	 deletion recorded in the Undo queue. The goal is to
	 allow procedure CellPasted to regenerate only one
	 column head when undoing the operation */
      TtaChangeInfoLastRegisteredElem (doc, 3);
      HandleColAndRowAlignAttributes (row, doc);
    }
}

/*----------------------------------------------------------------------
   CellPasted                                              
  ----------------------------------------------------------------------*/
void CellPasted (NotifyElement * event)
{
   Element             cell, row;
   ElementType         elType;
   Document            doc;
   AttributeType       rowspanType;
   Attribute           attr;
   int                 span;
   ThotBool            inMath;

   cell = event->element;
   doc = event->document;
   elType = TtaGetElementType (cell);
   inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
   row = TtaGetParent (cell);
   /* regenerate the corresponding ColumnHead except if it's called by
      undo for reinserting the cells deleted by a "Delete Column"
      command (only the first reinserted cell has to create a ColumnHead)
      See function RemoveColumn above */
   if (event->info == 4)
     /* undoing the deletion of the last cell in a "delete column"
	command. Regenerate the corresponding ColumnHead and link the
	restored cell with that ColumnHead, but do not generate empty
	cells in other rows */
     NewCell (cell, doc, TRUE, FALSE);
   else if (event->info == 3)
     {
     /* undoing the deletion of any other cell in a "delete column"
	command. Link the restored cell with the corresponding ColumnHead*/
       elType = TtaGetElementType (cell);
       inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
       LinkCellToColumnHead (cell, CurrentColumn, doc, inMath);
     }
   else
     /* undoing/redoing. Link the cell with ColumnHead elements, but do not
	generate empty cells in other rows */
     NewCell (cell, doc, FALSE, FALSE);

   /* update row extensions */
   rowspanType.AttrSSchema = elType.ElSSchema;
   if (inMath)
     rowspanType.AttrTypeNum = MathML_ATTR_rowspan_;
   else
     rowspanType.AttrTypeNum = HTML_ATTR_rowspan_;
   attr = TtaGetAttribute (cell, rowspanType);
   if (!attr)
     span = 1;
   else
     {
       span = TtaGetAttributeValue (attr);
       if (span < 0)
	 span = 1;
     }
   if (span > 1 || span == 0)
     SetRowExt (cell, span, doc, inMath);
   
   HandleColAndRowAlignAttributes (row, doc);
   /* Check attribute NAME or ID in order to make sure that its value */
   /* is unique in the document */
   MakeUniqueName (cell, doc);
}

/*----------------------------------------------------------------------
   UpdateRowspanForRow
   A new row has been created by the user (if addRow) or will be deleted
   (if not addRow). Take care of the rowspan attribute of the previous rows
   for creating cells in the row if it's a new row,
   or create empty cells in the next rows according of the rowspan attribute
   of the deleted row.
  ----------------------------------------------------------------------*/
static void UpdateRowspanForRow (Element row, Document doc, ThotBool inMath,
				 ThotBool addRow)
{
  Element             colhead, cell, prev, table;
  ElementType         elType;
  int                 colspan, i;

  elType = TtaGetElementType (row);
  if (inMath)
    elType.ElTypeNum = MathML_EL_MTABLE;
  else
    elType.ElTypeNum = HTML_EL_Table;
  table = TtaGetTypedAncestor (row, elType);
  /* get the first column */
  if (inMath)
    elType.ElTypeNum = MathML_EL_MColumn_head;
  else
    elType.ElTypeNum = HTML_EL_Column_head;
  colhead = TtaSearchTypedElement (elType, SearchInTree, table);
  /* check the cells for all columns */
  cell = NULL;
  while (colhead)
    {
      /* is there a cell in a row above that spans the current row? */
      prev = SpanningCellForRow (row, colhead, doc, inMath, TRUE, addRow,
				 &colspan);
      if (addRow && !prev)
	/* we are creating a new row and no cell from a row above is
           covering the current cell position. Create an empty cell at
	   that position */
	cell = AddEmptyCellInRow (row, colhead, cell, FALSE, doc, inMath,
				  FALSE, FALSE);
      else if (!addRow)
	/* we are deleting a row */
	{
	  if (prev)
	    /* a cell from a row above is covering the current cell position.
	       Create an empty cell to allow this row to be pasted correctly
	       later */
	    {
	      if (colspan > 0)
		for (i = 1; i <= colspan; i++)
		  cell = AddEmptyCellInRow (row, colhead, cell, FALSE, doc,
					    inMath, FALSE, TRUE);
	    }
	  else
	    cell = GetCellFromColumnHead (row, colhead, inMath);
	}
      if (colspan == 0)
	colhead = NULL;
      else
	while (colspan >= 1 && colhead)
	  {
	    TtaNextSibling (&colhead);
	    colspan--;
	  }
    }
}

/*----------------------------------------------------------------------
   DeleteRow
   A table row will be deleted by the user
  ----------------------------------------------------------------------*/
ThotBool DeleteRow (NotifyElement *event)
{
  Element             row, cell;
  ElementType         elType, cellType;
  Attribute           attr;
  AttributeType       rowspanType;
  Document            doc;
  int                 rowspan;
  ThotBool            inMath;

  row = event->element;
  doc = event->document;
  elType = TtaGetElementType (row);
  inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
  /* decrease the value of the rowspan attribute of cells from rows above
     that span the deleted row */
  UpdateRowspanForRow (row, doc, inMath, FALSE);
  /* create empty cells in the following rows, where cells of the deleted
     row span the following rows */
  rowspanType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    rowspanType.AttrTypeNum = MathML_ATTR_rowspan_;
  else
    rowspanType.AttrTypeNum = HTML_ATTR_rowspan_;
  cell = TtaGetFirstChild (row);
  while (cell)
    {
      cellType = TtaGetElementType (cell);
      if (cellType.ElSSchema == elType.ElSSchema &&
	  ((inMath && cellType.ElTypeNum == MathML_EL_MTD) ||
	   (!inMath && (cellType.ElTypeNum == HTML_EL_Data_cell ||
			cellType.ElTypeNum == HTML_EL_Heading_cell))))
	/* it's really a cell */
	{
	  /* check its rowspan attribute */
	  attr = TtaGetAttribute (cell, rowspanType);
	  if (attr)
	    /* this cell has an attribute rowspan */
	    {
	      rowspan = TtaGetAttributeValue (attr);
	      if (rowspan < 0)
		rowspan = 1;
	    }
	  else
	    rowspan = 1;
	  if (rowspan > 1 || rowspan == 0)
	    {
	      ChangeRowspan (cell, rowspan, 1, doc);
	      TtaRegisterAttributeDelete (attr, cell, doc);
	      TtaRemoveAttribute (cell, attr, doc);
	    }
	}
      TtaNextSibling (&cell);
    }
  return FALSE;		/* let Thot perform normal operation */
}

/*----------------------------------------------------------------------
   RowDeleted
   A table row has been deleted by the user
  ----------------------------------------------------------------------*/
void RowDeleted (NotifyElement *event)
{
  Element             rowgroup;
  Document            doc;

  doc = event->document;
  rowgroup = event->element;
  if (rowgroup == NULL)
    /* the rowgroup doesn't exist */
    return;
  HandleColAndRowAlignAttributes (rowgroup, doc);
}

/*----------------------------------------------------------------------
  DeleteColumn
  A column will be deleted by the user.
  ----------------------------------------------------------------------*/
ThotBool DeleteColumn (NotifyElement * event)
{
  Element             cell, prevCell, row, colhead, nextColhead, prev, table;
  Document            doc;
  ElementType         elType;
  Attribute           attr;
  AttributeType       attrTypeC, attrTypeR;
  int                 rowspan, colspan, rs;
  ThotBool            span, inMath;

      colhead = event->element;
      doc = event->document;
      elType = TtaGetElementType (colhead);
      elType.ElTypeNum = HTML_EL_Table;
      table = TtaGetTypedAncestor (colhead, elType);
      attrTypeC.AttrSSchema = elType.ElSSchema;
      attrTypeR.AttrSSchema = elType.ElSSchema;
      inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema("MathML",doc));
      if (inMath)
	{
	  elType.ElTypeNum = MathML_EL_TableRow;
	  attrTypeC.AttrTypeNum = MathML_ATTR_columnspan;
	  attrTypeR.AttrTypeNum = MathML_ATTR_rowspan_;
	}
      else
	{
	  elType.ElTypeNum = HTML_EL_Table_row;
	  attrTypeC.AttrTypeNum = HTML_ATTR_colspan_;
	  attrTypeR.AttrTypeNum = HTML_ATTR_rowspan_;
	}
      nextColhead = colhead;
      TtaNextSibling (&nextColhead);

      /* get the first row in the table */
      row = TtaSearchTypedElement (elType, SearchInTree, table);
      /* check all cells of the column */
      while (row)
	{
	  /* check if the cell has span values */
	  rowspan = 1;
	  colspan = 1;
	  cell = GetCellFromColumnHead (row, colhead, inMath);
	  if (cell)
	    {
	      attr = TtaGetAttribute (cell, attrTypeC);
	      if (attr)
		/* there is a colspan attribute */
		colspan = TtaGetAttributeValue (attr);
	      if (colspan > 1 || colspan == 0)
		{
		  ChangeColspan (cell, colspan, 1, doc);
		  TtaRegisterAttributeDelete (attr, cell, doc);
		  TtaRemoveAttribute (cell, attr, doc);
		}
	      attr = TtaGetAttribute (cell, attrTypeR);
	      if (attr)
		/* there is a rowspan */
		{
		  rowspan = TtaGetAttributeValue (attr);
		  if (rowspan < 0)
		    rowspan = 1;
		}
	    }
	  else
	    {
	      prevCell = GetCloseCell (row, colhead, doc, TRUE, FALSE,
				       inMath, &span, &rs, nextColhead==NULL);
	    }
	  prev = row;
	  if (rowspan == 0)
	    rowspan = THOT_MAXINT;
	  while (rowspan >= 1 && row)
	    {
	      row = GetSiblingRow (row, FALSE, inMath);
	      if (row)
		{
		  if (rowspan < THOT_MAXINT)
		    rowspan--;
		  if (!cell && rowspan >= 1)
		    prevCell = GetCloseCell (row, colhead, doc, TRUE, FALSE,
					     inMath, &span, &rs, FALSE);
		}
	    }
	  if (row == NULL)
	    {
	      prev = TtaGetParent (prev);
	      while (prev && row == NULL)
		{
		  TtaNextSibling (&prev);
		  row = TtaSearchTypedElementInTree (elType, SearchForward,
						     prev, prev);
		}
	    }
	}
  return FALSE;		/* let Thot perform normal operation */
}

/*----------------------------------------------------------------------
  ColumnPasted                                             
  ----------------------------------------------------------------------*/
void ColumnPasted (NotifyElement * event)
{
  Element         nextColhead, row, cell, table;
  ElementType     elType;
  Attribute       attr;
  AttributeType   attrTypeC;
  Document        doc;
  SSchema         tableSS;
  int             colspan;
  ThotBool        inMath;

  CurrentColumn = event->element;
  nextColhead = CurrentColumn;
  TtaNextSibling (&nextColhead);
  if (!nextColhead)
    /* we are pasting the last column of the table. Check the last
       cell of all rows. Those with an attribute colspan = 0 have
       to refer to the new colhead through their attribute ColExt. */
    {
      doc = event->document;
      elType = TtaGetElementType (CurrentColumn);
      tableSS = elType.ElSSchema;
      elType.ElTypeNum = HTML_EL_Table;
      table = TtaGetTypedAncestor (CurrentColumn, elType);
      attrTypeC.AttrSSchema = tableSS;      
      inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema("MathML",doc));
      if (inMath)
	{
	  elType.ElTypeNum = MathML_EL_TableRow;
	  attrTypeC.AttrTypeNum = MathML_ATTR_columnspan;
	}
      else
	{
	  elType.ElTypeNum = HTML_EL_Table_row;
	  attrTypeC.AttrTypeNum = HTML_ATTR_colspan_;
	}
      /* get the first row in the table */
      row = TtaSearchTypedElement (elType, SearchInTree, table);
      /* check all rows */
      while (row)
	{
	  cell = TtaGetLastChild (row);
	  /* process only cell elements */
	  if (cell)
	    elType = TtaGetElementType (cell);
	  while (cell &&
		 (elType.ElSSchema != tableSS ||
		  (inMath && elType.ElTypeNum != MathML_EL_MTD) ||
		  (!inMath && (elType.ElTypeNum != HTML_EL_Data_cell &&
			       elType.ElTypeNum != HTML_EL_Heading_cell))))
	    {
	      TtaPreviousSibling (&cell);
	      if (cell)
		elType = TtaGetElementType (cell);
	    }
	  if (cell)
	    /* this is the last cell in the row */
	    {
	      attr = TtaGetAttribute (cell, attrTypeC);
	      if (attr)
		/* there is a colspan attribute */
		{
		  colspan = TtaGetAttributeValue (attr);
		  if (colspan == 0)
		    /* this cell spans up to the last column */
		    SetColExt (cell, 0, doc, inMath, FALSE);
		}
	    }
	  row = GetSiblingRow (row, FALSE, inMath);
	}
    }
}

/*----------------------------------------------------------------------
  TablePasted                                             
  ----------------------------------------------------------------------*/
void TablePasted (NotifyElement * event)
{
  /* associate each cell with a column */
  CheckAllRows (event->element, event->document, FALSE, FALSE);
}

/*----------------------------------------------------------------------
  TableCreated                                            
  ----------------------------------------------------------------------*/
void TableCreated (NotifyElement * event)
{
   Element             table;
   Document            doc;
   ElementType         elType;
   AttributeType       attrType;
   Attribute           attr;

   table = event->element;
   doc = event->document;
   elType = TtaGetElementType (table);
   attrType.AttrSSchema = elType.ElSSchema;
   attrType.AttrTypeNum = HTML_ATTR_Border;
   attr = TtaGetAttribute (table, attrType);
   if (attr == NULL)
     {
	attr = TtaNewAttribute (attrType);
	if (attr != NULL)
	   TtaAttachAttribute (table, attr, doc);
     }
   TtaSetAttributeValue (attr, 1, table, doc);
   NewTable = TRUE;
}

/*----------------------------------------------------------------------
  CopyRow:
  A table row of has been copied. Generate empty cells where appropriate
  in the new copy.
  This function is called only by command Copy, not by command Cut.
  ----------------------------------------------------------------------*/
void CopyRow (Element copyRow, Element origRow, Document doc)
{
  Element       prevCopyCell, copyCell, origCell, colhead, table;
  ElementType   elType, cellType;
  Attribute     attr;
  AttributeType attrType;
  int           colspan;
  ThotBool      inMath;

  elType = TtaGetElementType (origRow);
  inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
  attrType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    {
      elType.ElTypeNum = MathML_EL_MTABLE;
      attrType.AttrTypeNum = MathML_ATTR_columnspan;
    }
  else
    {
      elType.ElTypeNum = HTML_EL_Table;
      attrType.AttrTypeNum = HTML_ATTR_colspan_;
    }
  /* get the table to which the original row belongs */
  table = TtaGetTypedAncestor (origRow, elType);
  /* get the first column head of this table */
  if (inMath)
    elType.ElTypeNum = MathML_EL_MColumn_head;
  else
    elType.ElTypeNum = HTML_EL_Column_head;
  colhead = TtaSearchTypedElement (elType, SearchInTree, table);
  /* get prepared for the generation of empty cells */
  cellType.ElSSchema = elType.ElSSchema;
  if (inMath)
    cellType.ElTypeNum = MathML_EL_MTD;
  else
    cellType.ElTypeNum = HTML_EL_Data_cell;

  /* check all column heads of the original table */
  prevCopyCell = NULL;
  while (colhead)
    {
      colspan = 1;
      /* is there a cell linked to that column head in the original row? */
      origCell = GetCellFromColumnHead (origRow, colhead, inMath);
      if (!origCell)
	/* no cell at that position in the original table */
	{
	  /* create an empty cell at the same position in the copy */
	  copyCell = TtaNewTree (doc, cellType, "");
	  if (copyCell)
	    {
	      if (prevCopyCell)
		TtaInsertSibling (copyCell, prevCopyCell, FALSE, doc);
	      else
		TtaInsertFirstChild (&copyCell, copyRow, doc);
	      prevCopyCell = copyCell;
	    }
	}
      else
	/* there is a cell at that position in the original row */
	{
	  if (prevCopyCell)
	    /* we have already processed some cells in the copy row.
	       skip one */
	    prevCopyCell = GetSiblingCell (prevCopyCell, FALSE, inMath);
	  else
	    /* we have not processed any cell in the copy row yet */
	    {
	      /* get the first cell in the copy row */
	      prevCopyCell = TtaGetFirstChild (copyRow);
	      if (prevCopyCell)
		elType = TtaGetElementType (prevCopyCell);
	      /* skip any other element until we reach the first cell */
	      while (prevCopyCell &&
		     (elType.ElSSchema != cellType.ElSSchema ||
		      (inMath && elType.ElTypeNum != MathML_EL_MTD) ||
		      (!inMath && (elType.ElTypeNum != HTML_EL_Data_cell &&
				   elType.ElTypeNum != HTML_EL_Heading_cell))))
		{
		  TtaNextSibling (&prevCopyCell);
		  if (prevCopyCell)
		    elType = TtaGetElementType (prevCopyCell);
		}
	    }
	  /* get the value of the colspan attribute for the current cell */
	  attr = TtaGetAttribute (origCell, attrType);
	  if (attr)
	    /* there is a colspan attribute */
	    {
	      colspan = TtaGetAttributeValue (attr);
	      if (colspan < 0)
		colspan = 1;
	    }
	}
      if (colspan == 0)
	/* this cell spans all the rest of the row. Stop */
	colhead = NULL;
      else
	/* skip a number of column heads equal to the spanning of the cell */
	while (colspan >= 1 && colhead)
	  {
	    TtaNextSibling (&colhead);
	    colspan--;
	  }
    }
}

/*----------------------------------------------------------------------
  CopyCell:
  a cell of a row or of a column has been copied.
  ----------------------------------------------------------------------*/
void CopyCell (Element cell, Document doc, ThotBool inRow)
{
  ElementType         elType;
  Attribute           attr;
  AttributeType       attrType, attrTypeR;
  ThotBool            inMath;
  
  elType = TtaGetElementType (cell);
  attrType.AttrSSchema = elType.ElSSchema;
  attrTypeR.AttrSSchema = elType.ElSSchema;
  inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
  if (inRow)
    {
      if (inMath)
	{
	  attrType.AttrTypeNum = MathML_ATTR_rowspan_;
	  attrTypeR.AttrTypeNum = MathML_ATTR_MRowExt;
	}
      else
	{
	  attrType.AttrTypeNum = HTML_ATTR_rowspan_;
	  attrTypeR.AttrTypeNum = HTML_ATTR_RowExt;
	}
      /* remove the rowspan attribute of the copy */
      attr = TtaGetAttribute (cell, attrType);
      if (attr)
	TtaRemoveAttribute (cell, attr, doc);
      /* remove the refext attribute of the copy */
      attr = TtaGetAttribute (cell, attrTypeR);
      if (attr)
	TtaRemoveAttribute (cell, attr, doc);
    }
  else
    {
      if (inMath)
	{
	  attrType.AttrTypeNum = MathML_ATTR_columnspan;
	  attrTypeR.AttrTypeNum = MathML_ATTR_MColExt;
	}
      else
	{
	  attrType.AttrTypeNum = HTML_ATTR_colspan_;
	  attrTypeR.AttrTypeNum = HTML_ATTR_ColExt;
	}
      /* remove the colspan attribute of the copy */
      attr = TtaGetAttribute (cell, attrType);
      if (attr)
	TtaRemoveAttribute (cell, attr, doc);
      /* remove the refext attribute of the copy */
      attr = TtaGetAttribute (cell, attrTypeR);
      if (attr)
	TtaRemoveAttribute (cell, attr, doc);
    }
}

/*----------------------------------------------------------------------
   TablebodyDeleted                                             
  ----------------------------------------------------------------------*/
void TablebodyDeleted (NotifyElement * event)
{
  Element             sibling, table;
  ElementType	      elType;
  Document            doc;
  ThotBool	      empty;

  if (event->info == 2)
     /* call from the Return key. Don't do anything */
     return;
  doc = event->document;
  table = event->element;
  sibling = TtaGetFirstChild (table);
  empty = TRUE;
  while (sibling != NULL && empty)
    {
      elType = TtaGetElementType (sibling);
      if (elType.ElTypeNum == HTML_EL_CAPTION ||
	  elType.ElTypeNum == HTML_EL_thead ||
	  elType.ElTypeNum == HTML_EL_Table_body ||
	  elType.ElTypeNum == HTML_EL_tfoot)
	if (TtaGetElementVolume (sibling) > 0)
	  empty = FALSE;
      TtaNextSibling (&sibling);
    }
  if (empty)
    {
      if (TtaPrepareUndo (doc))
	  /* register that the table is deleted */
	  TtaRegisterElementDelete (table, doc);
      TtaDeleteTree (table, doc);
    }
}

/*----------------------------------------------------------------------
   RowCreated                                              
  ----------------------------------------------------------------------*/
void RowCreated (NotifyElement *event)
{
  Element             row, table;
  Element             cell;
  ElementType         elType;
  Document            doc;
  ThotBool            inMath;

  row = event->element;
  doc = event->document;
  elType = TtaGetElementType (row);
  inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
  if (inMath)
    elType.ElTypeNum = MathML_EL_MTABLE;
  else
    elType.ElTypeNum = HTML_EL_Table;
  table = TtaGetTypedAncestor (row, elType);
  if (NewTable)
    {
      /* the table element is just created now
       We need to create the table_head element */
      if (inMath)
	CheckMTable (table, doc, FALSE);
      else
	CheckTable (table, doc);
      NewTable = FALSE;
    }
  else
    {
      /* avoid processing the pasted cells in the created row */
      CurrentRow = row;
      /* remove the cell created by the editor */
      cell = TtaGetFirstChild (row);
      if (cell)
	TtaDeleteTree (cell, doc);
      UpdateRowspanForRow (row, doc, inMath, TRUE);
    }
  HandleColAndRowAlignAttributes (row, doc);
  CurrentCell = TtaGetLastChild (row);
}

/*----------------------------------------------------------------------
   RowPasted                                               
  ----------------------------------------------------------------------*/
void RowPasted (NotifyElement * event)
{
  Element             row, table, colhead, prevCell, cell, nextCell,
                      prev;
  ElementType         elType, cellType;
  Attribute           attr;
  AttributeType       rowspanType, rowextType, colspanType;
  Document            doc;
  int                 colspan, span;
  ThotBool            inMath;

  row = event->element;
  doc = event->document;
  elType = TtaGetElementType (row);
  inMath = TtaSameSSchemas (elType.ElSSchema, TtaGetSSchema ("MathML", doc));
  if (inMath)
    elType.ElTypeNum = MathML_EL_MTABLE;
  else
    elType.ElTypeNum = HTML_EL_Table;
  table = TtaGetTypedAncestor (row, elType);

  /* only if we're not pasting a whole table */
  /* link each cell to the appropriate column head and add empty cells
     if the pasted row is too short */
  rowspanType.AttrSSchema = elType.ElSSchema;
  rowextType.AttrSSchema = elType.ElSSchema;
  colspanType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    {
      elType.ElTypeNum = MathML_EL_MColumn_head;
      rowspanType.AttrTypeNum = MathML_ATTR_rowspan_;
      rowextType.AttrTypeNum = MathML_ATTR_MRowExt;
      colspanType.AttrTypeNum = MathML_ATTR_columnspan;
    }
  else
    {
      elType.ElTypeNum = HTML_EL_Column_head;
      rowspanType.AttrTypeNum = HTML_ATTR_rowspan_;
      rowextType.AttrTypeNum = HTML_ATTR_RowExt;
      colspanType.AttrTypeNum = HTML_ATTR_colspan_;
    }
      /* get the first column */
  colhead = TtaSearchTypedElement (elType, SearchInTree, table);
  prevCell = NULL;
  /* get the first cell in the pasted row */
  cell = TtaGetFirstChild (row);
  cellType = TtaGetElementType (cell);
  if (cellType.ElSSchema != elType.ElSSchema ||
      (inMath && elType.ElTypeNum != MathML_EL_MTD) ||
      (!inMath && (cellType.ElTypeNum != HTML_EL_Data_cell &&
		   cellType.ElTypeNum != HTML_EL_Heading_cell)))
    cell = GetSiblingCell (cell, FALSE, inMath);
  while (colhead)
    {
      nextCell = GetSiblingCell (cell, FALSE, inMath);
      /* handle the colspan attribute */
      span = 1;
      if (cell)
	{
	  attr = TtaGetAttribute (cell, colspanType);
	  if (attr)
	    {
	      span = TtaGetAttributeValue (attr);
	      if (span < 0)
		span = 1;
	    }
	}
      /* is there a cell in a row above that spans this row? */
      prev = SpanningCellForRow (row, colhead, doc, inMath, TRUE, TRUE,
				 &colspan);
      if (!prev)
	/* no cell from a row above is covering the current cell position*/
	{
	  if (cell)
	    {
	      LinkCellToColumnHead (cell, colhead, doc, inMath);
	      if (span > 1 || span == 0)
		SetColExt (cell, span, doc, inMath, FALSE);
	    }
	  else
	    /* the pasted row has no cell for this column.
	       Add an empty cell */
	    cell = AddEmptyCellInRow (row, colhead, prevCell, FALSE, doc,
				      inMath, FALSE, FALSE);
	}
      else
	/* this position is taken by a cell from a row above */
	if (cell)
	  /* the pasted row has a cell at this position. Remove it */
	  {
	    TtaDeleteTree (cell, doc);
	    cell = NULL;
	  }
      /* get the next column where there is a free slot for the pasted
	 row */
      if (colspan == 0)
	colspan = THOT_MAXINT;
      if (span == 0)
	span = THOT_MAXINT;
      while ((colspan >= 1 || span >= 1) && colhead)
	{
	  TtaNextSibling (&colhead);
	  span--;
	  if (span == 0)
	    {
	      if (cell)
		prevCell = cell;
	      cell = nextCell;
	    }
	  colspan--;
	  if (colspan > 0)
	    {
	      nextCell = GetSiblingCell (cell, FALSE, inMath);
	      if (cell)
		TtaDeleteTree (cell, doc);
	    }
	}
    }
  if (nextCell)
    /* we have checked all columns of the table and there are extra cells.
       Remove them */
    while (nextCell)
      {
	cell = nextCell;
	nextCell = GetSiblingCell (cell, FALSE, inMath);
	TtaDeleteTree (cell, doc);
      }

  HandleColAndRowAlignAttributes (row, doc);
  /* avoid processing the cells of the created row */
  CurrentCell = TtaGetLastChild (row);
  /* Check attribute NAME or ID in order to make sure that its value */
  /* is unique in the document */
  MakeUniqueName (row, doc);
}

/*----------------------------------------------------------------------
   MoveCellContents
   Move all children of element nextCell as siblings of element previous
   (if it is not NULL) or as children of element cell (if previous is NULL),
   and delete element nextCell.
   When returning, previous is the latest element moved.
  ----------------------------------------------------------------------*/
static void MoveCellContents (Element nextCell, Element cell,
			      Element* previous, Document doc, ThotBool inMath)
{
  Element             child, nextChild;
  ElementType         elType;

  TtaRegisterElementDelete (nextCell, doc);
  if (TtaGetElementVolume (nextCell) == 0)
    /* empty cell */
    child = NULL;
  else
    {
      child = TtaGetFirstChild (nextCell);
      if (inMath)
	/* get the first element contained in the CellWrapper */
	child = TtaGetFirstChild (child);
    }
  /* move the contents of this cell to the cell whose attribute colspan
     has changed */
  while (child)
    {
      nextChild = child;  TtaNextSibling (&nextChild);
      TtaRemoveTree (child, doc);
      if (!inMath && *previous)
	/* if it's a pseudo-paragraph, change it into a
	   paragraph */
	{
	  elType = TtaGetElementType (child);
	  if (elType.ElTypeNum == HTML_EL_Pseudo_paragraph)
	    TtaChangeElementType (child, HTML_EL_Paragraph);
	}
      if (*previous)
	TtaInsertSibling (child, *previous, FALSE, doc);
      else
	TtaInsertFirstChild (&child, cell, doc);
      TtaRegisterElementCreate (child, doc);
      *previous = child;
      child = nextChild;
    }
  /* delete the cell */
  TtaDeleteTree (nextCell, doc);
}

/*----------------------------------------------------------------------
  NextCellInRow
  Return the cell that follows a given cell in the same row.
  ----------------------------------------------------------------------*/
Element NextCellInRow (Element cell, ThotBool inMath)
{
  return  GetSiblingCell (cell, FALSE, inMath);
}

/*----------------------------------------------------------------------
  GetActualColspan
  Return the number of columns comprised between the beginning of the cell
  and the end of the table.
  ----------------------------------------------------------------------*/
int GetActualColspan (Element cell, ThotBool inMath)
{
  Element            colhead;
  ElementType        elType;
  Attribute          attr;
  AttributeType      attrType;
  int                span;
  char               name[50];
  Document           refdoc;

  elType = TtaGetElementType (cell);
  attrType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    attrType.AttrTypeNum = MathML_ATTR_MRef_column;
  else
    attrType.AttrTypeNum = HTML_ATTR_Ref_column;
  attr = TtaGetAttribute (cell, attrType);
  if (attr)
    TtaGiveReferenceAttributeValue (attr, &colhead, name, &refdoc);
  span = 1;
  while (colhead)
    {
      TtaNextSibling (&colhead);
      if (colhead)
	span++;
    }
  return span;
}

/*----------------------------------------------------------------------
   ChangeColspan
   The value of the colspan attribute has changed from oldspan to newspan
   for the given cell. Add new empty cells (if newspan < oldspan) after
   element cell or merge the following cells (if oldspan > newspan).
  ----------------------------------------------------------------------*/
void ChangeColspan (Element cell, int oldspan, int newspan, Document doc)
{
  Element             table, nextCell, previous, colHead, curColHead, cHead,
                      row, curRow, prevCell;
  ElementType         tableType;
  AttributeType       attrType, rowspanType, colspanType;
  Attribute           attr;
  Document            refDoc;
  char                name[50];
  int                 ncol, nrow, i, j, rowspan, curColspan, curRowspan;
  ThotBool            inMath;

  if (oldspan == newspan)
    /* no change */
    return;
  if (cell == NULL)
    return;

  tableType = TtaGetElementType (cell);
  attrType.AttrSSchema = tableType.ElSSchema;
  rowspanType.AttrSSchema = tableType.ElSSchema;
  colspanType.AttrSSchema = tableType.ElSSchema;
  inMath = TtaSameSSchemas (tableType.ElSSchema, TtaGetSSchema ("MathML",doc));
  if (inMath)
    {
      tableType.ElTypeNum = MathML_EL_MTABLE;
      attrType.AttrTypeNum = MathML_ATTR_MRef_column;
      rowspanType.AttrTypeNum = MathML_ATTR_rowspan_;
      colspanType.AttrTypeNum = MathML_ATTR_columnspan;
    }
  else
    {
      tableType.ElTypeNum = HTML_EL_Table;
      attrType.AttrTypeNum = HTML_ATTR_Ref_column;
      rowspanType.AttrTypeNum = HTML_ATTR_rowspan_;
      colspanType.AttrTypeNum = HTML_ATTR_colspan_;
    }
  table = TtaGetTypedAncestor (cell, tableType);
  row = TtaGetParent (cell);

  /* get the rowspan value of the cell */
  attr = TtaGetAttribute (cell, rowspanType);
  if (attr)
    {
      rowspan = TtaGetAttributeValue (attr);
      if (rowspan < 0)
	rowspan = 1;
    }
  else
    rowspan = 1;
  /* get the column of the cell */
  attr = TtaGetAttribute (cell, attrType);
  if (attr)
    TtaGiveReferenceAttributeValue (attr, &colHead, name, &refDoc);
  else
    return;

  if ((newspan > oldspan && oldspan > 0) || newspan == 0)
    /* merge with cells in following columns */
    {
      previous = TtaGetLastChild (cell);
      if (inMath)
	/* get the last element contained in the CellWrapper */
	previous = TtaGetLastChild (previous);
      curColHead = colHead;
      ncol = 0;
      while (curColHead && (ncol < newspan || newspan == 0))
	{
	  ncol++; /* count columns */
	  if (ncol > oldspan)
	    {
	      curRow = row; nrow = 0;
	      while (curRow && (nrow < rowspan || rowspan == 0))
		{
		  nrow++;
		  nextCell = GetCellFromColumnHead (curRow, curColHead,inMath);
		  if (nextCell)
		    {
		      /* get the colspan value of the current cell */
		      attr = TtaGetAttribute (nextCell, colspanType);
		      if (attr)
			{
			  curColspan = TtaGetAttributeValue (attr);
			  if (curColspan < 0)
			    curColspan = 1;
			}
		      else
			curColspan = 1;
		      if (newspan > 0 &&
			  (curColspan > newspan - ncol + 1 || curColspan == 0))
			{
			  prevCell = nextCell;
			  cHead = colHead;
			  for (j = 0; j < newspan && cHead; j++)
			    TtaNextSibling (&cHead);
			  for (j = 0; j < curColspan-newspan+ncol-1 && cHead;
			       j++)
			    {
			      prevCell = AddEmptyCellInRow (curRow, cHead,
				    prevCell, FALSE, doc, inMath, FALSE, TRUE);
			      TtaNextSibling (&cHead);
			    }
			}
		      attr = TtaGetAttribute (nextCell, rowspanType);
		      if (attr)
			/* this cell has a rowspan attribute */
			{
			  curRowspan = TtaGetAttributeValue (attr);
			  if (curRowspan < 0)
			    curRowspan = 1;
			}
		      else
			curRowspan = 1;
		      if ((curRowspan == 0 || nrow + curRowspan -1 > rowspan)&&
			  rowspan > 0)
			ChangeRowspan (nextCell, curRowspan,
				       rowspan - nrow + 1, doc);
		      MoveCellContents (nextCell, cell, &previous, doc,inMath);
		    }
		  curRow = GetSiblingRow (curRow, FALSE, inMath);
		}
	    }
	  TtaNextSibling (&curColHead);
	}
    }
  else
    /* generate empty cells */
    {
      /* process all rows covered by the cell (due to attribute rowspan) */
      curRow = row; i = 0;
      while (curRow && (i < rowspan || rowspan == 0))
	{
	  i++;
	  if (curRow == row)
	    prevCell = cell;
	  else
	    {
	      cHead = colHead;
	      prevCell = NULL;
	      do
		{
		  TtaPreviousSibling (&cHead);
		  if (cHead)
		    prevCell = GetCellFromColumnHead (curRow, cHead,
						      inMath);
		}
	      while (cHead && !prevCell);
	    }
	  cHead = colHead;
	  /* skip the columns covered by the cell with its new colspan */
	  for (j = 0; j < newspan && cHead; j++)
	    TtaNextSibling (&cHead);
	  /* add new empty cells to fill the space left by reducing the
	     colspan value */
	  for (ncol = 0; (oldspan == 0 || ncol < oldspan - newspan) && cHead;
	       ncol++)
	    {
	      prevCell = AddEmptyCellInRow (curRow, cHead, prevCell, FALSE,
					    doc, inMath, FALSE, TRUE);
	      TtaNextSibling (&cHead);
	    }
	  /* next sibling of current row */
	  curRow = GetSiblingRow (curRow, FALSE, inMath);
	}
    }
}

/*----------------------------------------------------------------------
   ColspanCreated                                          
  ----------------------------------------------------------------------*/
void ColspanCreated (NotifyAttribute * event)
{
  Element             cell;
  Attribute           attr;
  Document            doc;
  int                 span, max, overflow;
  ThotBool            inMath;

  doc = event->document;
  cell = event->element;
  attr = event->attribute;
  inMath = TtaSameSSchemas (TtaGetElementType (cell).ElSSchema,
			    TtaGetSSchema ("MathML", doc));
  span = TtaGetAttributeValue (attr);
  if (span < 0 || span == 1)
    /* invalid value. Change it to 2 */
    {
      span = 2;
      TtaSetAttributeValue (attr, span, cell, doc);
    }
  else
    {
      max = GetActualColspan (cell, inMath);
      if (span > max)
	{
	  span = max;
	  TtaSetAttributeValue (attr, span, cell, doc);
	}
    }
  /* for undo operations just restore the col extension */
  if (event->info != 1)
    ChangeColspan (cell, 1, span, doc);
  overflow = SetColExt (cell, span, doc, inMath, FALSE);
}

/*----------------------------------------------------------------------
   RegisterColspan                                         
  ----------------------------------------------------------------------*/
ThotBool RegisterColspan (NotifyAttribute * event)
{
  PreviousColspan = TtaGetAttributeValue (event->attribute);
  if (PreviousColspan < 0)
    PreviousColspan = 1;
  return FALSE;		/* let Thot perform normal operation */
}

/*----------------------------------------------------------------------
   ColspanModified                                         
  ----------------------------------------------------------------------*/
void ColspanModified (NotifyAttribute * event)
{
  Element             cell;
  Attribute           attr;
  Document            doc;
  int                 span, max, overflow;
  ThotBool            inMath;

  doc = event->document;
  cell = event->element;
  attr = event->attribute;
  inMath = TtaSameSSchemas (TtaGetElementType (cell).ElSSchema,
			    TtaGetSSchema ("MathML", doc));
  span = TtaGetAttributeValue (attr);
  if (span == 1)
    /* keep the previous value to restore it when undoing the change */
    TtaSetAttributeValue (attr, PreviousColspan, cell, doc);
  else if (span < 0)
    /* invalid value, keep the previous value */
    {
      span = PreviousColspan;
      TtaSetAttributeValue (attr, span, cell, doc);
    }
  else
    {
      max = GetActualColspan (cell, inMath);
      if (span > max)
	{
	  span = max;
	  TtaSetAttributeValue (attr, span, cell, doc);
	}
    }
  if (span != PreviousColspan || span == 0)
    {
      /* for undo operations just restore the col extension */
      if (event->info != 1)
	ChangeColspan (cell, PreviousColspan, span, doc);
      overflow = SetColExt (cell, span, doc, inMath, FALSE);
    }
  if (span < 0 || span == 1)
    /* invalid value */
    {
      TtaRegisterAttributeDelete (attr, cell, doc);
      TtaRemoveAttribute (cell, attr, doc);
    }
}

/*----------------------------------------------------------------------
   ColspanDeleted
  ----------------------------------------------------------------------*/
void ColspanDeleted (NotifyAttribute * event)
{
  ElementType         elType;
  ThotBool            inMath;

  if (PreviousColspan > 1 || PreviousColspan == 0)
    {
      /* for undo operations just restore the col extension */
      if (event->info != 1)
	ChangeColspan (event->element, PreviousColspan, 1, event->document);
      elType = TtaGetElementType (event->element);
      inMath = TtaSameSSchemas (elType.ElSSchema,
				TtaGetSSchema ("MathML", event->document));
      SetColExt (event->element, 1, event->document, inMath, FALSE);
    }
}

/*----------------------------------------------------------------------
  NextCellInColumn
  Return the cell that follows a given cell in the same column.
  ----------------------------------------------------------------------*/
Element NextCellInColumn (Element cell, ThotBool inMath)
{
  Element            colhead, row, nextCell;
  ElementType        elType;
  Attribute          attr;
  AttributeType      attrType;
  char               name[50];
  Document           refdoc;

  nextCell = NULL;
  colhead = NULL;
  elType = TtaGetElementType (cell);
  attrType.AttrSSchema = elType.ElSSchema;
  if (inMath)
    attrType.AttrTypeNum = MathML_ATTR_MRef_column;
  else
    attrType.AttrTypeNum = HTML_ATTR_Ref_column;
  attr = TtaGetAttribute (cell, attrType);
  if (attr)
    TtaGiveReferenceAttributeValue (attr, &colhead, name, &refdoc);
  if (colhead)
    {
      row = TtaGetParent (cell);
      while (row && !nextCell)
	{
	  row = GetSiblingRow (row, FALSE, inMath);
	  if (row)
	    nextCell = GetCellFromColumnHead (row, colhead, inMath);
	}
    }
  return nextCell;
}

/*----------------------------------------------------------------------
  GetActualRowspan
  Return the number of rows comprised between the beginning of the cell
  and the end of the rows group.
  ----------------------------------------------------------------------*/
int GetActualRowspan (Element cell, ThotBool inMath)
{
  Element            row;
  int                span;

  row = TtaGetParent (cell);
  span = 1;
  while (row)
    {
      row = GetSiblingRow (row, FALSE, inMath);
      if (row)
	span++;
    }
  return span;
}

/*----------------------------------------------------------------------
   ChangeRowspan
   The value of the rowspan attribute has changed from oldspan to newspan
   for the given cell. Add new empty cells (if newspan < oldspan) below
   the cell or merge the cells below (if oldspan > newspan).
  ----------------------------------------------------------------------*/
void ChangeRowspan (Element cell, int oldspan, int newspan, Document doc)
{
  Element             table, row, previous, nextCell, colHead, curColHead,
                      prevCell, siblingColhead;
  ElementType         tableType;
  AttributeType       attrType, colspanType, rowspanType;
  Attribute           attr;
  Document            refDoc;
  char                name[50];
  int                 i, nrows, colspan, curColspan, curRowspan;
  ThotBool            inMath, before;

  if (oldspan == newspan)
    /* no change */
    return;
  if (cell == NULL)
    return;

  tableType = TtaGetElementType (cell);
  attrType.AttrSSchema = tableType.ElSSchema;
  colspanType.AttrSSchema = tableType.ElSSchema;
  rowspanType.AttrSSchema = tableType.ElSSchema;
  inMath = TtaSameSSchemas (tableType.ElSSchema, TtaGetSSchema ("MathML",doc));
  if (inMath)
    {
      tableType.ElTypeNum = MathML_EL_MTABLE;
      attrType.AttrTypeNum = MathML_ATTR_MRef_column;
      colspanType.AttrTypeNum = MathML_ATTR_columnspan;
      rowspanType.AttrTypeNum = MathML_ATTR_rowspan_;
    }
  else
    {
      tableType.ElTypeNum = HTML_EL_Table;
      attrType.AttrTypeNum = HTML_ATTR_Ref_column;
      colspanType.AttrTypeNum = HTML_ATTR_colspan_;
      rowspanType.AttrTypeNum = HTML_ATTR_rowspan_;
    }
  table = TtaGetTypedAncestor (cell, tableType);
  row = TtaGetParent (cell);

  /* get the colspan value of the cell */
  attr = TtaGetAttribute (cell, colspanType);
  if (attr)
    {
      colspan = TtaGetAttributeValue (attr);
      if (colspan < 0)
	colspan = 1;
    }
  else
    colspan = 1;
  /* get the column of the cell */
  attr = TtaGetAttribute (cell, attrType);
  if (attr)
    TtaGiveReferenceAttributeValue (attr, &colHead, name, &refDoc);
  else
    return;

  nrows = 0;
  if ((newspan > oldspan && oldspan > 0) || newspan == 0)
    /* merge with following cells */
    {
      previous = TtaGetLastChild (cell);
      if (inMath)
	/* get the last element contained in the CellWrapper */
	previous = TtaGetLastChild (previous);
      while (row && (newspan == 0 || nrows < newspan))
	{
	  nrows++; /* count rows */
	  if (nrows > oldspan)
	    {
	      nextCell = GetCellFromColumnHead (row, colHead, inMath);
	      if (nextCell)
		{
		  attr = TtaGetAttribute (nextCell, colspanType);
		  if (attr)
		    /* this cell has a colspan attribute */
		    {
		      curColspan = TtaGetAttributeValue (attr);
		      if (curColspan < 0)
			curColspan = 1;
		      else if (curColspan == 0)
			curColspan = THOT_MAXINT;
		    }
		  else
		    curColspan = 1;
		  curColHead = colHead;
		  if (curColspan > colspan && colspan > 0)
		    /* create empty cells */
		    {
		      prevCell = nextCell;
		      for (i = 1; i < curColspan && curColHead; i++)
			{
			  TtaNextSibling (&curColHead);
			  if (i >= colspan && curColHead)
			    prevCell = AddEmptyCellInRow (row, curColHead,
							  prevCell, FALSE, doc,
							  inMath, FALSE, TRUE);
			}
		    }
		  else if (colspan > 0)
		    /* merge cells from that row */
		    for (i = curColspan; i <= colspan && curColHead; i++)
		      {
			if (nextCell)
			  {
			    attr = TtaGetAttribute (nextCell, rowspanType);
			    if (attr)
			      /* this cell has a rowspan attribute */
			      {
				curRowspan = TtaGetAttributeValue (attr);
				if (curRowspan < 0)
				  curRowspan = 1;
			      }
			    else
			      curRowspan = 1;
			    if ((curRowspan == 0 ||
				 nrows + curRowspan - 1 > newspan) &&
				newspan>0)
			      /* this cell extends beyound the limit of the
				 extension of the cell of interest. Fill the
			         holes with empty cells */
			      ChangeRowspan (nextCell, curRowspan,
					     newspan - nrows + 1, doc);
			    MoveCellContents (nextCell, cell, &previous, doc,
					      inMath);
			  }
			TtaNextSibling (&curColHead);
			nextCell = GetCellFromColumnHead (row, curColHead,
							  inMath);
		      }
		}
	    }
	  row = GetSiblingRow (row, FALSE, inMath);
	}
    }
  else if ((newspan < oldspan && newspan > 0) || oldspan == 0)
    /* generate empty cells */
    {
      while (row && (nrows < oldspan || oldspan == 0))
	{
	  nrows++;
	  if (nrows > newspan)
	    {
	      prevCell = NULL;
	      siblingColhead = colHead;
	      TtaPreviousSibling (&siblingColhead);
	      if (siblingColhead)
		{
		  prevCell = GetCellFromColumnHead (row, siblingColhead,
						    inMath);
		  if (prevCell)
		    before = FALSE;
		}
	      if (!prevCell)
		{
		  before = TRUE;
		  siblingColhead = colHead;
		  while (siblingColhead && !prevCell)
		    {
		      TtaNextSibling (&siblingColhead);
		      prevCell = GetCellFromColumnHead (row,siblingColhead,
							inMath);
		    }
		}
	      curColHead = colHead;
	      for (i = 0; (i < colspan || colspan == 0) && curColHead; i++)
		{
		  prevCell = AddEmptyCellInRow (row, curColHead, prevCell,
					     before, doc, inMath, FALSE, TRUE);
		  before = FALSE;
		  TtaNextSibling (&curColHead);
		}
	    }
	  row = GetSiblingRow (row, FALSE, inMath);
	}
    }
}

/*----------------------------------------------------------------------
   RowspanCreated                                          
  ----------------------------------------------------------------------*/
void RowspanCreated (NotifyAttribute * event)
{
  Element             cell;
  Attribute           attr;
  Document            doc;
  int                 span, max;
  ThotBool            inMath;

  doc = event->document;
  cell = event->element;
  attr = event->attribute;
  inMath = TtaSameSSchemas (TtaGetElementType (cell).ElSSchema,
			    TtaGetSSchema ("MathML", doc));
  span = TtaGetAttributeValue (attr);
  if (span < 0 || span == 1)
    /* invalid value. Change it to 2 */
    {
      span = 2;
      TtaSetAttributeValue (attr, span, cell, doc);
    }
  else
    {
      max = GetActualRowspan (cell, inMath);
      if (span > max)
	{
	  span = max;
	  TtaSetAttributeValue (attr, span, cell, doc);
	}
    }
  /* for undo operations just restore the row extension */
  if (event->info != 1)
    {
      /* the creation of the attribute must be registered after the creation
	 of the empty cells generated by ChangeRowspan.*************/
      ChangeRowspan (cell, 1, span, doc);
    }
  SetRowExt (cell, span, doc, inMath);
}

/*----------------------------------------------------------------------
   RegisterRowspan                                         
  ----------------------------------------------------------------------*/
ThotBool RegisterRowspan (NotifyAttribute * event)
{
  PreviousRowspan = TtaGetAttributeValue (event->attribute);
  if (PreviousRowspan < 0)
    PreviousRowspan = 1;
  return FALSE;		/* let Thot perform normal operation */
}

/*----------------------------------------------------------------------
   RowspanModified                                         
  ----------------------------------------------------------------------*/
void RowspanModified (NotifyAttribute * event)
{
  Element             cell;
  Attribute           attr;
  Document            doc;
  int                 span, max;
  ThotBool            inMath;

  doc = event->document;
  cell = event->element;
  attr = event->attribute;
  inMath = TtaSameSSchemas (TtaGetElementType (cell).ElSSchema,
			    TtaGetSSchema ("MathML", doc));
  span = TtaGetAttributeValue (attr);
  if (span == 1)
    /* keep the previous value to restore it when undoing the change */
    TtaSetAttributeValue (attr, PreviousRowspan, cell, doc);
  else if (span < 0)
    /* invalid value, keep the previous value */
    {
      span = PreviousRowspan;
      TtaSetAttributeValue (attr, span, cell, doc);
    }
  else
    {
      max = GetActualRowspan (cell, inMath);
      if (span > max)
	{
	  span = max;
	  TtaSetAttributeValue (attr, span, cell, doc);
	}
    }
  if (span != PreviousRowspan)
    {
      /* for undo operations just restore the row extension */
      if (event->info != 1)
	  /* ********************* the old value of the attribute should be registered after the
	     creation of the empty cells generated by ChangeRowspan. */
	  ChangeRowspan (cell, PreviousRowspan, span, doc);
      SetRowExt (cell, span, doc, inMath);
    }
  if (span < 0 || span == 1)
    /* invalid value */
    {
      TtaRegisterAttributeDelete (attr, cell, doc);
      TtaRemoveAttribute (cell, attr, doc);
    }
}

/*----------------------------------------------------------------------
   RowspanDeleted
  ----------------------------------------------------------------------*/
void RowspanDeleted (NotifyAttribute * event)
{
  ElementType         elType;
  ThotBool            inMath;

  if (PreviousRowspan > 1 || PreviousRowspan == 0)
    {
      /* for undo operations just restore the row extension */
      if (event->info != 1)
	ChangeRowspan (event->element, PreviousRowspan, 1, event->document);
      elType = TtaGetElementType (event->element);
      inMath = TtaSameSSchemas (elType.ElSSchema,
				TtaGetSSchema ("MathML", event->document));
      SetRowExt (event->element, 1, event->document, inMath);
    }
}
